#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def check_resource_counts(count_map):
    for name, count in count_map.items():
        test.assertEqual(count,
                         len(list(reality.resources_by_logical_name(name))))


example_template = Template({
    'A': RsrcDef({'!a': 'initial'}, []),
    'B': RsrcDef({'!b': 'first'}, ['A']),
})
engine.create_stack('foo', example_template)
engine.noop(4)
engine.call(verify, example_template)

example_template_inverted = Template({
    'A': RsrcDef({'!a': 'updated'}, ['B']),
    'B': RsrcDef({'!b': 'second'}, []),
})
engine.update_stack('foo', example_template_inverted)
engine.noop(4)
engine.call(check_resource_counts, {'A': 2, 'B': 1})
engine.noop(2)
engine.call(verify, example_template_inverted)
engine.call(check_resource_counts, {'A': 1, 'B': 1})

engine.delete_stack('foo')
engine.noop(3)
engine.call(verify, Template({}))



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""ResourcePropertiesData object."""

from oslo_config import cfg
from oslo_serialization import jsonutils
from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from heat.common import crypt
from heat.db import api as db_api
from heat.objects import fields as heat_fields


class ResourcePropertiesData(
    base.VersionedObject,
    base.VersionedObjectDictCompat,
    base.ComparableVersionedObject,
):
    fields = {
        'id': fields.IntegerField(),
        'data': heat_fields.JsonField(nullable=True),
        'created_at': fields.DateTimeField(read_only=True),
        'updated_at': fields.DateTimeField(nullable=True),
    }

    @staticmethod
    def _from_db_object(rpd, context, db_rpd, data_unencrypted=None):
        # The data_unencrypted field allows us to avoid an extra
        # decrypt operation, e.g. when called from create().
        for field in rpd.fields:
            rpd[field] = db_rpd[field]
        if data_unencrypted:  # save a little (decryption) processing
            rpd['data'] = data_unencrypted
        elif db_rpd['encrypted'] and rpd['data'] is not None:
            rpd['data'] = crypt.decrypted_dict(rpd['data'])

        # TODO(cwolfe) setting the context here should go away, that
        # should have been done with the initialisation of the rpd
        # object. For now, maintaining consistency with other
        # _from_db_object methods.
        rpd._context = context
        rpd.obj_reset_changes()
        return rpd

    @classmethod
    def create_or_update(cls, context, data, rpd_id=None):
        properties_data_encrypted, properties_data = \
            ResourcePropertiesData.encrypt_properties_data(data)
        values = {'encrypted': properties_data_encrypted,
                  'data': properties_data}
        db_obj = db_api.resource_prop_data_create_or_update(
            context, values, rpd_id)
        return cls._from_db_object(cls(), context, db_obj, data)

    @classmethod
    def create(cls, context, data):
        return ResourcePropertiesData.create_or_update(context, data)

    @staticmethod
    def encrypt_properties_data(data):
        if cfg.CONF.encrypt_parameters_and_properties and data:
            result = {}
            for prop_name, prop_value in data.items():
                prop_string = jsonutils.dumps(prop_value)
                encrypted_value = crypt.encrypt(prop_string)
                result[prop_name] = encrypted_value
            return (True, result)
        return (False, data)

    @staticmethod
    def get_by_id(context, id):
        db_ref = db_api.resource_prop_data_get(context, id)
        return ResourcePropertiesData._from_db_object(
            ResourcePropertiesData(context=context), context, db_ref)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime
from unittest import mock
import uuid

from oslo_config import cfg
from oslo_messaging.rpc import dispatcher
from oslo_serialization import jsonutils as json
from oslo_utils import timeutils

from heat.common import crypt
from heat.common import exception
from heat.common import template_format
from heat.db import api as db_api
from heat.engine.clients.os import swift
from heat.engine.clients.os import zaqar
from heat.engine import service
from heat.engine import service_software_config
from heat.engine import software_config_io as swc_io
from heat.objects import resource as resource_objects
from heat.objects import software_config as software_config_object
from heat.objects import software_deployment as software_deployment_object
from heat.tests import common
from heat.tests.engine import tools
from heat.tests import utils

software_config_inputs = '''
heat_template_version: 2013-05-23
description: Validate software config input/output types

resources:

  InputOutputTestConfig:
    type: OS::Heat::SoftwareConfig
    properties:
      group: puppet
      inputs:
      - name: boolean_input
        type: Boolean
      - name: json_input
        type: Json
      - name: number_input
        type: Number
      - name: string_input
        type: String
      - name: comma_delimited_list_input
        type: CommaDelimitedList
      outputs:
      - name: boolean_output
        type: Boolean
      - name: json_output
        type: Json
      - name: number_output
        type: Number
      - name: string_output
        type: String
      - name: comma_delimited_list_output
        type: CommaDelimitedList

'''


class SoftwareConfigServiceTest(common.HeatTestCase):

    def setUp(self):
        super(SoftwareConfigServiceTest, self).setUp()
        self.ctx = utils.dummy_context()
        self.engine = service.EngineService('a-host', 'a-topic')

    def _create_software_config(
            self, group='Heat::Shell', name='config_mysql', config=None,
            inputs=None, outputs=None, options=None, context=None):
        cntx = context if context else self.ctx
        inputs = inputs or []
        outputs = outputs or []
        options = options or {}
        return self.engine.create_software_config(
            cntx, group, name, config, inputs, outputs, options)

    def _create_dummy_config_object(self):
        obj_config = software_config_object.SoftwareConfig()
        obj_config['id'] = str(uuid.uuid4())
        obj_config['name'] = 'myconfig'
        obj_config['group'] = 'mygroup'
        obj_config['config'] = {'config': 'hello world',
                                'inputs': [],
                                'outputs': [],
                                'options': {}}
        obj_config['created_at'] = timeutils.utcnow()
        return obj_config

    def assert_status_reason(self, expected, actual):
        expected_dict = dict((i.split(' : ') for i in expected.split(', ')))
        actual_dict = dict((i.split(' : ') for i in actual.split(', ')))
        self.assertEqual(expected_dict, actual_dict)

    def test_list_software_configs(self):
        config = self._create_software_config()
        self.assertIsNotNone(config)
        config_id = config['id']

        configs = self.engine.list_software_configs(self.ctx)
        self.assertIsNotNone(configs)
        config_ids = [x['id'] for x in configs]
        self.assertIn(config_id, config_ids)

        admin_cntx = utils.dummy_context(is_admin=True)

        admin_config = self._create_software_config(context=admin_cntx)
        admin_config_id = admin_config['id']
        configs = self.engine.list_software_configs(admin_cntx)
        self.assertIsNotNone(configs)
        config_ids = [x['id'] for x in configs]
        project_ids = [x['project'] for x in configs]
        self.assertEqual(2, len(project_ids))
        self.assertEqual(2, len(config_ids))
        self.assertIn(config_id, config_ids)
        self.assertIn(admin_config_id, config_ids)

    def test_show_software_config(self):
        config_id = str(uuid.uuid4())

        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.engine.show_software_config,
                               self.ctx, config_id)
        self.assertEqual(exception.NotFound, ex.exc_info[0])

        config = self._create_software_config()
        res = self.engine.show_software_config(self.ctx, config['id'])
        self.assertEqual(config, res)

    def test_create_software_config_new_ids(self):
        config1 = self._create_software_config()
        self.assertIsNotNone(config1)

        config2 = self._create_software_config()
        self.assertNotEqual(config1['id'], config2['id'])

    def test_create_software_config(self):
        kwargs = {
            'group': 'Heat::Chef',
            'name': 'config_heat',
            'config': '...',
            'inputs': [{'name': 'mode'}],
            'outputs': [{'name': 'endpoint'}],
            'options': {}
        }
        config = self._create_software_config(**kwargs)
        config_id = config['id']
        config = self.engine.show_software_config(self.ctx, config_id)
        self.assertEqual(kwargs['group'], config['group'])
        self.assertEqual(kwargs['name'], config['name'])
        self.assertEqual(kwargs['config'], config['config'])
        self.assertEqual([{'name': 'mode', 'type': 'String'}],
                         config['inputs'])
        self.assertEqual([{'name': 'endpoint', 'type': 'String',
                           'error_output': False}],
                         config['outputs'])
        self.assertEqual(kwargs['options'], config['options'])

    def test_create_config_exceeds_max_per_tenant(self):
        cfg.CONF.set_override('max_software_configs_per_tenant', 0)
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self._create_software_config)
        self.assertEqual(exception.RequestLimitExceeded, ex.exc_info[0])
        self.assertIn("You have reached the maximum software configs "
                      "per tenant", str(ex.exc_info[1]))

    def test_create_software_config_structured(self):
        kwargs = {
            'group': 'json-file',
            'name': 'config_heat',
            'config': {'foo': 'bar'},
            'inputs': [{'name': 'mode'}],
            'outputs': [{'name': 'endpoint'}],
            'options': {}
        }
        config = self._create_software_config(**kwargs)
        config_id = config['id']
        config = self.engine.show_software_config(self.ctx, config_id)
        self.assertEqual(kwargs['config'], config['config'])

    def test_delete_software_config(self):
        config = self._create_software_config()
        self.assertIsNotNone(config)
        config_id = config['id']
        self.engine.delete_software_config(self.ctx, config_id)

        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.engine.show_software_config,
                               self.ctx, config_id)
        self.assertEqual(exception.NotFound, ex.exc_info[0])

    def test_boolean_inputs_valid(self):
        stack_name = 'test_boolean_inputs_valid'
        t = template_format.parse(software_config_inputs)
        stack = utils.parse_stack(t, stack_name=stack_name)
        try:
            stack.validate()
        except exception.StackValidationFailed as exc:
            self.fail("Validation should have passed: %s" % str(exc))

    def _create_software_deployment(self, config_id=None, input_values=None,
                                    action='INIT',
                                    status='COMPLETE', status_reason='',
                                    config_group=None,
                                    server_id=str(uuid.uuid4()),
                                    config_name=None,
                                    stack_user_project_id=None):
        input_values = input_values or {}
        if config_id is None:
            config = self._create_software_config(group=config_group,
                                                  name=config_name)
            config_id = config['id']
        return self.engine.create_software_deployment(
            self.ctx, server_id, config_id, input_values,
            action, status, status_reason, stack_user_project_id)

    def test_list_software_deployments(self):
        stack_name = 'test_list_software_deployments'
        t = template_format.parse(tools.wp_template)
        stack = utils.parse_stack(t, stack_name=stack_name)

        fc = tools.setup_mocks_with_mock(self, stack)
        stack.store()
        stack.create()
        server = stack['WebServer']
        server_id = server.resource_id
        deployment = self._create_software_deployment(
            server_id=server_id)
        deployment_id = deployment['id']
        self.assertIsNotNone(deployment)

        deployments = self.engine.list_software_deployments(
            self.ctx, server_id=None)
        self.assertIsNotNone(deployments)
        deployment_ids = [x['id'] for x in deployments]
        self.assertIn(deployment_id, deployment_ids)
        self.assertIn(deployment, deployments)

        deployments = self.engine.list_software_deployments(
            self.ctx, server_id=str(uuid.uuid4()))
        self.assertEqual([], deployments)

        deployments = self.engine.list_software_deployments(
            self.ctx, server_id=server.resource_id)
        self.assertEqual([deployment], deployments)

        rsrcs = resource_objects.Resource.get_all_by_physical_resource_id(
            self.ctx, server_id)
        self.assertEqual(deployment['config_id'],
                         rsrcs[0].rsrc_metadata.get('deployments')[0]['id'])
        tools.validate_setup_mocks_with_mock(stack, fc)

    def test_metadata_software_deployments(self):
        stack_name = 'test_metadata_software_deployments'
        t = template_format.parse(tools.wp_template)
        stack = utils.parse_stack(t, stack_name=stack_name)

        fc = tools.setup_mocks_with_mock(self, stack)
        stack.store()
        stack.create()
        server = stack['WebServer']
        server_id = server.resource_id

        stack_user_project_id = str(uuid.uuid4())
        d1 = self._create_software_deployment(
            config_group='mygroup',
            server_id=server_id,
            config_name='02_second',
            stack_user_project_id=stack_user_project_id)
        d2 = self._create_software_deployment(
            config_group='mygroup',
            server_id=server_id,
            config_name='01_first',
            stack_user_project_id=stack_user_project_id)
        d3 = self._create_software_deployment(
            config_group='myothergroup',
            server_id=server_id,
            config_name='03_third',
            stack_user_project_id=stack_user_project_id)
        metadata = self.engine.metadata_software_deployments(
            self.ctx, server_id=server_id)
        self.assertEqual(3, len(metadata))
        self.assertEqual('mygroup', metadata[1]['group'])
        self.assertEqual('mygroup', metadata[0]['group'])
        self.assertEqual('myothergroup', metadata[2]['group'])
        self.assertEqual(d1['config_id'], metadata[1]['id'])
        self.assertEqual(d2['config_id'], metadata[0]['id'])
        self.assertEqual(d3['config_id'], metadata[2]['id'])
        self.assertEqual('01_first', metadata[0]['name'])
        self.assertEqual('02_second', metadata[1]['name'])
        self.assertEqual('03_third', metadata[2]['name'])

        # assert that metadata via metadata_software_deployments matches
        # metadata via server resource
        rsrcs = resource_objects.Resource.get_all_by_physical_resource_id(
            self.ctx, server_id)
        self.assertEqual(metadata, rsrcs[0].rsrc_metadata.get('deployments'))

        deployments = self.engine.metadata_software_deployments(
            self.ctx, server_id=str(uuid.uuid4()))
        self.assertEqual([], deployments)

        # assert get results when the context tenant_id matches
        # the stored stack_user_project_id
        ctx = utils.dummy_context(tenant_id=stack_user_project_id)
        metadata = self.engine.metadata_software_deployments(
            ctx, server_id=server_id)
        self.assertEqual(3, len(metadata))

        # assert get no results when the context tenant_id is unknown
        ctx = utils.dummy_context(tenant_id=str(uuid.uuid4()))
        metadata = self.engine.metadata_software_deployments(
            ctx, server_id=server_id)
        self.assertEqual(0, len(metadata))

        # assert None config is filtered out
        obj_conf = self._create_dummy_config_object()
        side_effect = [obj_conf, obj_conf, None]
        self.patchobject(software_config_object.SoftwareConfig,
                         '_from_db_object', side_effect=side_effect)
        metadata = self.engine.metadata_software_deployments(
            self.ctx, server_id=server_id)
        self.assertEqual(2, len(metadata))
        tools.validate_setup_mocks_with_mock(stack, fc)

    def test_show_software_deployment(self):
        deployment_id = str(uuid.uuid4())
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.engine.show_software_deployment,
                               self.ctx, deployment_id)
        self.assertEqual(exception.NotFound, ex.exc_info[0])

        deployment = self._create_software_deployment()
        self.assertIsNotNone(deployment)
        deployment_id = deployment['id']
        self.assertEqual(
            deployment,
            self.engine.show_software_deployment(self.ctx, deployment_id))

    def test_check_software_deployment(self):
        deployment_id = str(uuid.uuid4())
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.engine.check_software_deployment,
                               self.ctx, deployment_id, 10)
        self.assertEqual(exception.NotFound, ex.exc_info[0])

        deployment = self._create_software_deployment()
        self.assertIsNotNone(deployment)
        deployment_id = deployment['id']
        self.assertEqual(
            deployment,
            self.engine.check_software_deployment(self.ctx, deployment_id, 10))

    @mock.patch.object(service_software_config.SoftwareConfigService,
                       '_push_metadata_software_deployments')
    def test_signal_software_deployment(self, pmsd):
        self.assertRaises(ValueError,
                          self.engine.signal_software_deployment,
                          self.ctx, None, {}, None)
        deployment_id = str(uuid.uuid4())
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.engine.signal_software_deployment,
                               self.ctx, deployment_id, {}, None)
        self.assertEqual(exception.NotFound, ex.exc_info[0])

        deployment = self._create_software_deployment()
        deployment_id = deployment['id']

        # signal is ignore unless deployment is IN_PROGRESS
        self.assertIsNone(self.engine.signal_software_deployment(
            self.ctx, deployment_id, {}, None))

        # simple signal, no data
        deployment = self._create_software_deployment(action='INIT',
                                                      status='IN_PROGRESS')
        deployment_id = deployment['id']
        res = self.engine.signal_software_deployment(
            self.ctx, deployment_id, {}, None)
        self.assertEqual('deployment %s succeeded' % deployment_id, res)

        sd = software_deployment_object.SoftwareDeployment.get_by_id(
            self.ctx, deployment_id)
        self.assertEqual('COMPLETE', sd.status)
        self.assertEqual('Outputs received', sd.status_reason)
        self.assertEqual({
            'deploy_status_code': None,
            'deploy_stderr': None,
            'deploy_stdout': None
        }, sd.output_values)
        self.assertIsNotNone(sd.updated_at)

        # simple signal, some data
        config = self._create_software_config(outputs=[{'name': 'foo'}])
        deployment = self._create_software_deployment(
            config_id=config['id'], action='INIT', status='IN_PROGRESS')
        deployment_id = deployment['id']
        result = self.engine.signal_software_deployment(
            self.ctx,
            deployment_id,
            {'foo': 'bar', 'deploy_status_code': 0},
            None)
        self.assertEqual('deployment %s succeeded' % deployment_id, result)
        sd = software_deployment_object.SoftwareDeployment.get_by_id(
            self.ctx, deployment_id)
        self.assertEqual('COMPLETE', sd.status)
        self.assertEqual('Outputs received', sd.status_reason)
        self.assertEqual({
            'deploy_status_code': 0,
            'foo': 'bar',
            'deploy_stderr': None,
            'deploy_stdout': None
        }, sd.output_values)
        self.assertIsNotNone(sd.updated_at)

        # failed signal on deploy_status_code
        config = self._create_software_config(outputs=[{'name': 'foo'}])
        deployment = self._create_software_deployment(
            config_id=config['id'], action='INIT', status='IN_PROGRESS')
        deployment_id = deployment['id']
        result = self.engine.signal_software_deployment(
            self.ctx,
            deployment_id,
            {
                'foo': 'bar',
                'deploy_status_code': -1,
                'deploy_stderr': 'Its gone Pete Tong'
            },
            None)
        self.assertEqual('deployment %s failed (-1)' % deployment_id, result)
        sd = software_deployment_object.SoftwareDeployment.get_by_id(
            self.ctx, deployment_id)
        self.assertEqual('FAILED', sd.status)
        self.assert_status_reason(
            ('deploy_status_code : Deployment exited with non-zero '
             'status code: -1'),
            sd.status_reason)
        self.assertEqual({
            'deploy_status_code': -1,
            'foo': 'bar',
            'deploy_stderr': 'Its gone Pete Tong',
            'deploy_stdout': None
        }, sd.output_values)
        self.assertIsNotNone(sd.updated_at)

        # failed signal on error_output foo
        config = self._create_software_config(outputs=[
            {'name': 'foo', 'error_output': True}])
        deployment = self._create_software_deployment(
            config_id=config['id'], action='INIT', status='IN_PROGRESS')
        deployment_id = deployment['id']
        result = self.engine.signal_software_deployment(
            self.ctx,
            deployment_id,
            {
                'foo': 'bar',
                'deploy_status_code': -1,
                'deploy_stderr': 'Its gone Pete Tong'
            },
            None)
        self.assertEqual('deployment %s failed' % deployment_id, result)

        sd = software_deployment_object.SoftwareDeployment.get_by_id(
            self.ctx, deployment_id)
        self.assertEqual('FAILED', sd.status)
        self.assert_status_reason(
            ('foo : bar, deploy_status_code : Deployment exited with '
             'non-zero status code: -1'),
            sd.status_reason)
        self.assertEqual({
            'deploy_status_code': -1,
            'foo': 'bar',
            'deploy_stderr': 'Its gone Pete Tong',
            'deploy_stdout': None
        }, sd.output_values)
        self.assertIsNotNone(sd.updated_at)

    def test_create_software_deployment(self):
        kwargs = {
            'group': 'Heat::Chef',
            'name': 'config_heat',
            'config': '...',
            'inputs': [{'name': 'mode'}],
            'outputs': [{'name': 'endpoint'}],
            'options': {}
        }
        config = self._create_software_config(**kwargs)
        config_id = config['id']
        kwargs = {
            'config_id': config_id,
            'input_values': {'mode': 'standalone'},
            'action': 'INIT',
            'status': 'COMPLETE',
            'status_reason': ''
        }
        deployment = self._create_software_deployment(**kwargs)
        deployment_id = deployment['id']
        deployment = self.engine.show_software_deployment(
            self.ctx, deployment_id)
        self.assertEqual(deployment_id, deployment['id'])
        self.assertEqual(kwargs['input_values'], deployment['input_values'])

    def test_create_deployment_exceeds_max_per_tenant(self):
        cfg.CONF.set_override('max_software_deployments_per_tenant', 0)
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self._create_software_deployment)
        self.assertEqual(exception.RequestLimitExceeded, ex.exc_info[0])
        self.assertIn("You have reached the maximum software deployments"
                      " per tenant", str(ex.exc_info[1]))

    def test_create_software_deployment_invalid_stack_user_project_id(self):
        sc_kwargs = {
            'group': 'Heat::Chef',
            'name': 'config_heat',
            'config': '...',
            'inputs': [{'name': 'mode'}],
            'outputs': [{'name': 'endpoint'}],
            'options': {}
        }
        config = self._create_software_config(**sc_kwargs)
        config_id = config['id']
        sd_kwargs = {
            'config_id': config_id,
            'input_values': {'mode': 'standalone'},
            'action': 'INIT',
            'status': 'COMPLETE',
            'status_reason': '',
            # stack_user_project should be no more than 64 characters
            'stack_user_project_id': 'a' * 65
        }
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self._create_software_deployment,
                               **sd_kwargs)
        self.assertEqual(exception.Invalid, ex.exc_info[0])

    @mock.patch.object(service_software_config.SoftwareConfigService,
                       '_refresh_swift_software_deployment')
    def test_show_software_deployment_refresh(
            self, _refresh_swift_software_deployment):
        temp_url = ('http://192.0.2.1/v1/AUTH_a/b/c'
                    '?temp_url_sig=ctemp_url_expires=1234')
        config = self._create_software_config(inputs=[
            {
                'name': 'deploy_signal_transport',
                'type': 'String',
                'value': 'TEMP_URL_SIGNAL'
            }, {
                'name': 'deploy_signal_id',
                'type': 'String',
                'value': temp_url
            }
        ])

        deployment = self._create_software_deployment(
            status='IN_PROGRESS', config_id=config['id'])

        deployment_id = deployment['id']
        sd = software_deployment_object.SoftwareDeployment.get_by_id(
            self.ctx, deployment_id)
        _refresh_swift_software_deployment.return_value = sd
        self.assertEqual(
            deployment,
            self.engine.show_software_deployment(self.ctx, deployment_id))
        self.assertEqual(
            (self.ctx, sd, temp_url),
            _refresh_swift_software_deployment.call_args[0])

    def test_update_software_deployment_new_config(self):

        server_id = str(uuid.uuid4())
        mock_push = self.patchobject(self.engine.software_config,
                                     '_push_metadata_software_deployments')

        deployment = self._create_software_deployment(server_id=server_id)
        self.assertIsNotNone(deployment)
        deployment_id = deployment['id']
        deployment_action = deployment['action']
        self.assertEqual('INIT', deployment_action)
        config_id = deployment['config_id']
        self.assertIsNotNone(config_id)
        updated = self.engine.update_software_deployment(
            self.ctx, deployment_id=deployment_id, config_id=config_id,
            input_values={}, output_values={}, action='DEPLOY',
            status='WAITING', status_reason='', updated_at=None)
        self.assertIsNotNone(updated)
        self.assertEqual(config_id, updated['config_id'])
        self.assertEqual('DEPLOY', updated['action'])
        self.assertEqual('WAITING', updated['status'])
        self.assertEqual(2, mock_push.call_count)

    def test_update_software_deployment_status(self):

        server_id = str(uuid.uuid4())
        mock_push = self.patchobject(self.engine.software_config,
                                     '_push_metadata_software_deployments')

        deployment = self._create_software_deployment(server_id=server_id)

        self.assertIsNotNone(deployment)
        deployment_id = deployment['id']
        deployment_action = deployment['action']
        self.assertEqual('INIT', deployment_action)
        updated = self.engine.update_software_deployment(
            self.ctx, deployment_id=deployment_id, config_id=None,
            input_values=None, output_values={}, action='DEPLOY',
            status='WAITING', status_reason='', updated_at=None)
        self.assertIsNotNone(updated)
        self.assertEqual('DEPLOY', updated['action'])
        self.assertEqual('WAITING', updated['status'])

        mock_push.assert_called_once_with(self.ctx, server_id, None)

    def test_update_software_deployment_fields(self):

        deployment = self._create_software_deployment()
        deployment_id = deployment['id']
        config_id = deployment['config_id']

        def check_software_deployment_updated(**kwargs):
            values = {
                'config_id': None,
                'input_values': {},
                'output_values': {},
                'action': {},
                'status': 'WAITING',
                'status_reason': ''
            }
            values.update(kwargs)
            updated = self.engine.update_software_deployment(
                self.ctx, deployment_id, updated_at=None, **values)
            for key, value in kwargs.items():
                self.assertEqual(value, updated[key])

        check_software_deployment_updated(config_id=config_id)
        check_software_deployment_updated(input_values={'foo': 'fooooo'})
        check_software_deployment_updated(output_values={'bar': 'baaaaa'})
        check_software_deployment_updated(action='DEPLOY')
        check_software_deployment_updated(status='COMPLETE')
        check_software_deployment_updated(status_reason='Done!')

    @mock.patch.object(service_software_config.SoftwareConfigService,
                       '_push_metadata_software_deployments')
    def test_delete_software_deployment(self, pmsd):
        deployment_id = str(uuid.uuid4())
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self.engine.delete_software_deployment,
                               self.ctx, deployment_id)
        self.assertEqual(exception.NotFound, ex.exc_info[0])

        deployment = self._create_software_deployment()
        self.assertIsNotNone(deployment)
        deployment_id = deployment['id']

        deployments = self.engine.list_software_deployments(
            self.ctx, server_id=None)
        deployment_ids = [x['id'] for x in deployments]
        self.assertIn(deployment_id, deployment_ids)
        self.engine.delete_software_deployment(self.ctx, deployment_id)

        # assert one call for the create, and one for the delete
        pmsd.assert_has_calls([
            mock.call(self.ctx, deployment['server_id'], None),
            mock.call(self.ctx, deployment['server_id'], None)
        ])

        deployments = self.engine.list_software_deployments(
            self.ctx, server_id=None)
        deployment_ids = [x['id'] for x in deployments]
        self.assertNotIn(deployment_id, deployment_ids)

    @mock.patch.object(service_software_config.SoftwareConfigService,
                       'metadata_software_deployments')
    @mock.patch.object(db_api, 'resource_update')
    @mock.patch.object(db_api, 'resource_get_by_physical_resource_id')
    @mock.patch.object(service_software_config.requests, 'put')
    def test_push_metadata_software_deployments(
            self, put, res_get, res_upd, md_sd):
        rs = mock.Mock()
        rs.rsrc_metadata = {'original': 'metadata'}
        rs.id = '1234'
        rs.atomic_key = 1
        rs.data = []
        res_get.return_value = rs
        res_upd.return_value = 1

        deployments = {'deploy': 'this'}
        md_sd.return_value = deployments

        result_metadata = {
            'original': 'metadata',
            'deployments': {'deploy': 'this'}
        }

        self.engine.software_config._push_metadata_software_deployments(
            self.ctx, '1234', None)

        res_upd.assert_called_once_with(
            self.ctx, '1234', {'rsrc_metadata': result_metadata}, 1)
        put.side_effect = Exception('Unexpected requests.put')

    @mock.patch.object(service_software_config.SoftwareConfigService,
                       'metadata_software_deployments')
    @mock.patch.object(db_api, 'resource_update')
    @mock.patch.object(db_api, 'resource_get_by_physical_resource_id')
    @mock.patch.object(service_software_config.requests, 'put')
    def test_push_metadata_software_deployments_retry(
            self, put, res_get, res_upd, md_sd):
        rs = mock.Mock()
        rs.rsrc_metadata = {'original': 'metadata'}
        rs.id = '1234'
        rs.atomic_key = 1
        rs.data = []
        res_get.return_value = rs
        # zero update means another transaction updated
        res_upd.return_value = 0

        deployments = {'deploy': 'this'}
        md_sd.return_value = deployments

        f = self.engine.software_config._push_metadata_software_deployments
        self.patchobject(f.retry, 'sleep')
        self.assertRaises(
            exception.ConcurrentTransaction,
            f,
            self.ctx,
            '1234',
            None)

        # retry ten times then the final failure
        self.assertEqual(11, res_upd.call_count)
        put.assert_not_called()

    @mock.patch.object(service_software_config.SoftwareConfigService,
                       'metadata_software_deployments')
    @mock.patch.object(db_api, 'resource_update')
    @mock.patch.object(db_api, 'resource_get_by_physical_resource_id')
    @mock.patch.object(service_software_config.requests, 'put')
    def test_push_metadata_software_deployments_temp_url(
            self, put, res_get, res_upd, md_sd):
        rs = mock.Mock()
        rs.rsrc_metadata = {'original': 'metadata'}
        rs.id = '1234'
        rs.atomic_key = 1
        rd = mock.Mock()
        rd.key = 'metadata_put_url'
        rd.value = 'http://192.168.2.2/foo/bar'
        rs.data = [rd]
        res_get.return_value = rs
        res_upd.return_value = 1

        deployments = {'deploy': 'this'}
        md_sd.return_value = deployments

        result_metadata = {
            'original': 'metadata',
            'deployments': {'deploy': 'this'}
        }
        self.engine.software_config._push_metadata_software_deployments(
            self.ctx, '1234', None)
        res_upd.assert_has_calls([
            mock.call(self.ctx, '1234',
                      {'rsrc_metadata': result_metadata}, 1),
            mock.call(self.ctx, '1234',
                      {}, 2),
        ])

        put.assert_called_once_with(
            'http://192.168.2.2/foo/bar', json.dumps(result_metadata),
            timeout=60)

    @mock.patch.object(service_software_config.SoftwareConfigService,
                       'metadata_software_deployments')
    @mock.patch.object(db_api, 'resource_update')
    @mock.patch.object(db_api, 'resource_get_by_physical_resource_id')
    @mock.patch.object(zaqar.ZaqarClientPlugin, 'create_for_tenant')
    def test_push_metadata_software_deployments_queue(
            self, plugin, res_get, res_upd, md_sd):
        rs = mock.Mock()
        rs.rsrc_metadata = {'original': 'metadata'}
        rs.id = '1234'
        rs.atomic_key = 1
        rd = mock.Mock()
        rd.key = 'metadata_queue_id'
        rd.value = '6789'
        rs.data = [rd]
        res_get.return_value = rs
        res_upd.return_value = 1
        queue = mock.Mock()
        zaqar_client = mock.Mock()
        plugin.return_value = zaqar_client
        zaqar_client.queue.return_value = queue

        deployments = {'deploy': 'this'}
        md_sd.return_value = deployments

        result_metadata = {
            'original': 'metadata',
            'deployments': {'deploy': 'this'}
        }

        self.engine.software_config._push_metadata_software_deployments(
            self.ctx, '1234', 'project1')
        res_upd.assert_called_once_with(
            self.ctx, '1234', {'rsrc_metadata': result_metadata}, 1)

        plugin.assert_called_once_with('project1', mock.ANY)
        zaqar_client.queue.assert_called_once_with('6789')
        queue.post.assert_called_once_with(
            {'body': result_metadata, 'ttl': 3600})

    @mock.patch.object(service_software_config.SoftwareConfigService,
                       'signal_software_deployment')
    @mock.patch.object(swift.SwiftClientPlugin, '_create')
    def test_refresh_swift_software_deployment(self, scc, ssd):
        temp_url = ('http://192.0.2.1/v1/AUTH_a/b/c'
                    '?temp_url_sig=ctemp_url_expires=1234')
        container = 'b'
        object_name = 'c'

        config = self._create_software_config(inputs=[
            {
                'name': 'deploy_signal_transport',
                'type': 'String',
                'value': 'TEMP_URL_SIGNAL'
            }, {
                'name': 'deploy_signal_id',
                'type': 'String',
                'value': temp_url
            }
        ])

        timeutils.set_time_override(
            datetime.datetime(2013, 1, 23, 22, 48, 5, 0))
        self.addCleanup(timeutils.clear_time_override)
        now = timeutils.utcnow()
        then = now - datetime.timedelta(0, 60)

        last_modified_1 = 'Wed, 23 Jan 2013 22:47:05 GMT'
        last_modified_2 = 'Wed, 23 Jan 2013 22:48:05 GMT'

        sc = mock.MagicMock()
        headers = {
            'last-modified': last_modified_1
        }
        sc.head_object.return_value = headers
        sc.get_object.return_value = (headers, '{"foo": "bar"}')
        scc.return_value = sc

        deployment = self._create_software_deployment(
            status='IN_PROGRESS', config_id=config['id'])

        deployment_id = str(deployment['id'])
        sd = software_deployment_object.SoftwareDeployment.get_by_id(
            self.ctx, deployment_id)

        # poll with missing object
        swift_exc = swift.SwiftClientPlugin.exceptions_module
        sc.head_object.side_effect = swift_exc.ClientException(
            'Not found', http_status=404)

        self.assertEqual(
            sd,
            self.engine.software_config._refresh_swift_software_deployment(
                self.ctx, sd, temp_url))
        sc.head_object.assert_called_once_with(container, object_name)
        # no call to get_object or signal_last_modified
        self.assertEqual([], sc.get_object.mock_calls)
        self.assertEqual([], ssd.mock_calls)

        # poll with other error
        sc.head_object.side_effect = swift_exc.ClientException(
            'Ouch', http_status=409)
        self.assertRaises(
            swift_exc.ClientException,
            self.engine.software_config._refresh_swift_software_deployment,
            self.ctx,
            sd,
            temp_url)
        # no call to get_object or signal_last_modified
        self.assertEqual([], sc.get_object.mock_calls)
        self.assertEqual([], ssd.mock_calls)
        sc.head_object.side_effect = None

        # first poll populates data signal_last_modified
        self.engine.software_config._refresh_swift_software_deployment(
            self.ctx, sd, temp_url)
        sc.head_object.assert_called_with(container, object_name)
        sc.get_object.assert_called_once_with(container, object_name)
        # signal_software_deployment called with signal
        ssd.assert_called_once_with(self.ctx, deployment_id, {"foo": "bar"},
                                    then.isoformat())

        # second poll updated_at populated with first poll last-modified
        software_deployment_object.SoftwareDeployment.update_by_id(
            self.ctx, deployment_id, {'updated_at': then})
        sd = software_deployment_object.SoftwareDeployment.get_by_id(
            self.ctx, deployment_id)
        self.assertEqual(then, sd.updated_at)
        self.engine.software_config._refresh_swift_software_deployment(
            self.ctx, sd, temp_url)
        sc.get_object.assert_called_once_with(container, object_name)
        # signal_software_deployment has not been called again
        ssd.assert_called_once_with(self.ctx, deployment_id, {"foo": "bar"},
                                    then.isoformat())

        # third poll last-modified changed, new signal
        headers['last-modified'] = last_modified_2
        sc.head_object.return_value = headers
        sc.get_object.return_value = (headers, '{"bar": "baz"}')
        self.engine.software_config._refresh_swift_software_deployment(
            self.ctx, sd, temp_url)

        # two calls to signal_software_deployment, for then and now
        self.assertEqual(2, len(ssd.mock_calls))
        ssd.assert_called_with(self.ctx, deployment_id, {"bar": "baz"},
                               now.isoformat())

        # four polls result in only two signals, for then and now
        software_deployment_object.SoftwareDeployment.update_by_id(
            self.ctx, deployment_id, {'updated_at': now})
        sd = software_deployment_object.SoftwareDeployment.get_by_id(
            self.ctx, deployment_id)
        self.engine.software_config._refresh_swift_software_deployment(
            self.ctx, sd, temp_url)
        self.assertEqual(2, len(ssd.mock_calls))

    @mock.patch.object(service_software_config.SoftwareConfigService,
                       'signal_software_deployment')
    @mock.patch.object(service_software_config.SoftwareConfigService,
                       'metadata_software_deployments')
    @mock.patch.object(db_api, 'resource_update')
    @mock.patch.object(db_api, 'resource_get_by_physical_resource_id')
    @mock.patch.object(zaqar.ZaqarClientPlugin, 'create_for_tenant')
    def test_refresh_zaqar_software_deployment(self, plugin, res_get, res_upd,
                                               md_sd, ssd):
        rs = mock.Mock()
        rs.rsrc_metadata = {}
        rs.id = '1234'
        rs.atomic_key = 1
        rd1 = mock.Mock()
        rd1.key = 'user'
        rd1.value = 'user1'
        rd2 = mock.Mock()
        rd2.key = 'password'
        rd2.decrypt_method, rd2.value = crypt.encrypt('pass1')
        rs.data = [rd1, rd2]
        res_get.return_value = rs

        res_upd.return_value = 1
        deployments = {'deploy': 'this'}
        md_sd.return_value = deployments
        config = self._create_software_config(inputs=[
            {
                'name': 'deploy_signal_transport',
                'type': 'String',
                'value': 'ZAQAR_SIGNAL'
            }, {
                'name': 'deploy_queue_id',
                'type': 'String',
                'value': '6789'
            }
        ])

        queue = mock.Mock()
        zaqar_client = mock.Mock()
        plugin.return_value = zaqar_client
        zaqar_client.queue.return_value = queue
        queue.pop.return_value = [mock.Mock(body='ok')]

        deployment = self._create_software_deployment(
            status='IN_PROGRESS', config_id=config['id'])

        deployment_id = deployment['id']
        self.assertEqual(
            deployment,
            self.engine.show_software_deployment(self.ctx, deployment_id))

        zaqar_client.queue.assert_called_once_with('6789')
        queue.pop.assert_called_once_with()
        ssd.assert_called_once_with(self.ctx, deployment_id, 'ok', None)


class SoftwareConfigIOSchemaTest(common.HeatTestCase):
    def test_input_config_empty(self):
        name = 'foo'
        inp = swc_io.InputConfig(name=name)
        self.assertIsNone(inp.default())
        self.assertIs(False, inp.replace_on_change())
        self.assertEqual(name, inp.name())
        self.assertEqual({'name': name, 'type': 'String'}, inp.as_dict())
        self.assertEqual((name, None), inp.input_data())

    def test_input_config(self):
        name = 'bar'
        inp = swc_io.InputConfig(name=name, description='test', type='Number',
                                 default=0, replace_on_change=True)
        self.assertEqual(0, inp.default())
        self.assertIs(True, inp.replace_on_change())
        self.assertEqual(name, inp.name())
        self.assertEqual({'name': name, 'type': 'Number',
                          'description': 'test', 'default': 0,
                          'replace_on_change': True},
                         inp.as_dict())
        self.assertEqual((name, None), inp.input_data())

    def test_input_config_value(self):
        name = 'baz'
        inp = swc_io.InputConfig(name=name, type='Number',
                                 default=0, value=42)
        self.assertEqual(0, inp.default())
        self.assertIs(False, inp.replace_on_change())
        self.assertEqual(name, inp.name())
        self.assertEqual({'name': name, 'type': 'Number',
                          'default': 0, 'value': 42},
                         inp.as_dict())
        self.assertEqual((name, 42), inp.input_data())

    def test_input_config_no_name(self):
        self.assertRaises(ValueError, swc_io.InputConfig, type='String')

    def test_input_config_extra_key(self):
        self.assertRaises(ValueError, swc_io.InputConfig,
                          name='test', bogus='wat')

    def test_input_types(self):
        swc_io.InputConfig(name='str', type='String').as_dict()
        swc_io.InputConfig(name='num', type='Number').as_dict()
        swc_io.InputConfig(name='list', type='CommaDelimitedList').as_dict()
        swc_io.InputConfig(name='json', type='Json').as_dict()
        swc_io.InputConfig(name='bool', type='Boolean').as_dict()

        self.assertRaises(ValueError, swc_io.InputConfig,
                          name='bogus', type='BogusType')

    def test_output_config_empty(self):
        name = 'foo'
        outp = swc_io.OutputConfig(name=name)
        self.assertEqual(name, outp.name())
        self.assertEqual({'name': name, 'type': 'String',
                          'error_output': False},
                         outp.as_dict())

    def test_output_config(self):
        name = 'bar'
        outp = swc_io.OutputConfig(name=name, description='test',
                                   type='Json', error_output=True)
        self.assertEqual(name, outp.name())
        self.assertIs(True, outp.error_output())
        self.assertEqual({'name': name, 'type': 'Json',
                          'description': 'test', 'error_output': True},
                         outp.as_dict())

    def test_output_config_no_name(self):
        self.assertRaises(ValueError, swc_io.OutputConfig, type='String')

    def test_output_config_extra_key(self):
        self.assertRaises(ValueError, swc_io.OutputConfig,
                          name='test', bogus='wat')

    def test_output_types(self):
        swc_io.OutputConfig(name='str', type='String').as_dict()
        swc_io.OutputConfig(name='num', type='Number').as_dict()
        swc_io.OutputConfig(name='list', type='CommaDelimitedList').as_dict()
        swc_io.OutputConfig(name='json', type='Json').as_dict()
        swc_io.OutputConfig(name='bool', type='Boolean').as_dict()

        self.assertRaises(ValueError, swc_io.OutputConfig,
                          name='bogus', type='BogusType')

    def test_check_io_schema_empty_list(self):
        swc_io.check_io_schema_list([])

    def test_check_io_schema_string(self):
        self.assertRaises(TypeError, swc_io.check_io_schema_list, '')

    def test_check_io_schema_dict(self):
        self.assertRaises(TypeError, swc_io.check_io_schema_list, {})

    def test_check_io_schema_list_dict(self):
        swc_io.check_io_schema_list([{'name': 'foo'}])

    def test_check_io_schema_list_string(self):
        self.assertRaises(TypeError, swc_io.check_io_schema_list, ['foo'])

    def test_check_io_schema_list_list(self):
        self.assertRaises(TypeError, swc_io.check_io_schema_list, [['foo']])

    def test_check_io_schema_list_none(self):
        self.assertRaises(TypeError, swc_io.check_io_schema_list, [None])

    def test_check_io_schema_list_mixed(self):
        self.assertRaises(TypeError, swc_io.check_io_schema_list,
                          [{'name': 'foo'}, ('name', 'bar')])

    def test_input_config_value_json_default(self):
        name = 'baz'
        inp = swc_io.InputConfig(name=name, type='Json',
                                 default={'a': 1}, value=42)
        self.assertEqual({'a': 1}, inp.default())

    def test_input_config_value_default_coerce(self):
        name = 'baz'
        inp = swc_io.InputConfig(name=name, type='Number',
                                 default='0')
        self.assertEqual(0, inp.default())

    def test_input_config_value_ignore_string(self):
        name = 'baz'
        inp = swc_io.InputConfig(name=name, type='Number',
                                 default='')

        self.assertEqual({'type': 'Number', 'name': 'baz', 'default': ''},
                         inp.as_dict())


class SoftwareConfigServiceTestWithConstraint(SoftwareConfigServiceTest):
    """Test cases which require FK constraints"""

    def setUp(self):
        super(SoftwareConfigServiceTestWithConstraint, self).setUp()
        self.useFixture(utils.ForeignKeyConstraintFixture())

    def test_create_software_deployment_invalid_config_id(self):
        kwargs = {
            'config_id': 'this_is_invalid',
            'input_values': {'mode': 'standalone'},
            'action': 'INIT',
            'status': 'COMPLETE',
            'status_reason': ''
        }
        ex = self.assertRaises(dispatcher.ExpectedException,
                               self._create_software_deployment,
                               **kwargs)
        self.assertEqual(exception.Invalid, ex.exc_info[0])



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from heat.engine import translation


class ManilaShareNetwork(resource.Resource):
    """A resource that stores network information for share servers.

    Stores network information that will be used by share servers,
    where shares are hosted.
    """

    support_status = support.SupportStatus(version='5.0.0')

    PROPERTIES = (
        NAME, NEUTRON_NETWORK, NEUTRON_SUBNET, NOVA_NETWORK,
        DESCRIPTION, SECURITY_SERVICES,
    ) = (
        'name', 'neutron_network', 'neutron_subnet', 'nova_network',
        'description', 'security_services',
    )

    ATTRIBUTES = (
        SEGMENTATION_ID, CIDR, IP_VERSION, NETWORK_TYPE,
    ) = (
        'segmentation_id', 'cidr', 'ip_version', 'network_type',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of the share network.'),
            update_allowed=True
        ),
        NEUTRON_NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('Neutron network id.'),
            update_allowed=True,
            constraints=[constraints.CustomConstraint('neutron.network')]
        ),
        NEUTRON_SUBNET: properties.Schema(
            properties.Schema.STRING,
            _('Neutron subnet id.'),
            update_allowed=True,
            constraints=[constraints.CustomConstraint('neutron.subnet')]
        ),
        NOVA_NETWORK: properties.Schema(
            properties.Schema.STRING,
            _('Nova network id.'),
            update_allowed=True,
            support_status=support.SupportStatus(
                status=support.HIDDEN,
                message=_('nova-network is no longer supported.'),
                version='21.0.0'
            )
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Share network description.'),
            update_allowed=True
        ),
        SECURITY_SERVICES: properties.Schema(
            properties.Schema.LIST,
            _('A list of security services IDs or names.'),
            schema=properties.Schema(
                properties.Schema.STRING
            ),
            update_allowed=True,
            default=[]
        )
    }

    attributes_schema = {
        SEGMENTATION_ID: attributes.Schema(
            _('VLAN ID for VLAN networks or tunnel-id for GRE/VXLAN '
              'networks.'),
            type=attributes.Schema.STRING
        ),
        CIDR: attributes.Schema(
            _('CIDR of subnet.'),
            type=attributes.Schema.STRING
        ),
        IP_VERSION: attributes.Schema(
            _('Version of IP address.'),
            type=attributes.Schema.STRING
        ),
        NETWORK_TYPE: attributes.Schema(
            _('The physical mechanism by which the virtual network is '
              'implemented.'),
            type=attributes.Schema.STRING
        ),
    }

    default_client_name = 'manila'

    entity = 'share_networks'

    def _request_network(self):
        return self.client().share_networks.get(self.resource_id)

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        network = self._request_network()
        return getattr(network, name, None)

    def validate(self):
        super(ManilaShareNetwork, self).validate()

        if self.properties[self.NOVA_NETWORK]:
            msg = _('Nova network is no longer supported')
            raise exception.StackValidationFailed(message=msg)

        if (self.properties[self.NEUTRON_NETWORK] and not
                self.properties[self.NEUTRON_SUBNET]):
            raise exception.ResourcePropertyDependency(
                prop1=self.NEUTRON_NETWORK, prop2=self.NEUTRON_SUBNET)

        if (self.properties[self.NEUTRON_NETWORK] and
                self.properties[self.NEUTRON_SUBNET]):
            plg = self.client_plugin('neutron')
            subnet_id = plg.find_resourceid_by_name_or_id(
                plg.RES_TYPE_SUBNET, self.properties[self.NEUTRON_SUBNET])
            net_id = plg.network_id_from_subnet_id(subnet_id)
            provided_net_id = plg.find_resourceid_by_name_or_id(
                plg.RES_TYPE_NETWORK, self.properties[self.NEUTRON_NETWORK])
            if net_id != provided_net_id:
                msg = (_('Provided %(subnet)s does not belong '
                         'to provided %(network)s.')
                       % {'subnet': self.NEUTRON_SUBNET,
                          'network': self.NEUTRON_NETWORK})
                raise exception.StackValidationFailed(message=msg)

    def translation_rules(self, props):
        neutron_client_plugin = self.client_plugin('neutron')
        translation_rules = [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.NEUTRON_NETWORK],
                client_plugin=neutron_client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=neutron_client_plugin.RES_TYPE_NETWORK
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.NEUTRON_SUBNET],
                client_plugin=neutron_client_plugin,
                finder='find_resourceid_by_name_or_id',
                entity=neutron_client_plugin.RES_TYPE_SUBNET
            )
        ]
        return translation_rules

    def handle_create(self):
        neutron_subnet_id = self.properties[self.NEUTRON_SUBNET]
        neutron_net_id = self.properties[self.NEUTRON_NETWORK]
        if neutron_subnet_id and not neutron_net_id:
            neutron_net_id = self.client_plugin(
                'neutron').network_id_from_subnet_id(neutron_subnet_id)
        network = self.client().share_networks.create(
            name=self.properties[self.NAME],
            neutron_net_id=neutron_net_id,
            neutron_subnet_id=neutron_subnet_id,
            description=self.properties[self.DESCRIPTION])
        self.resource_id_set(network.id)

        for service in self.properties.get(self.SECURITY_SERVICES):
            self.client().share_networks.add_security_service(
                self.resource_id,
                self.client_plugin().get_security_service(service).id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if self.SECURITY_SERVICES in prop_diff:
            services = prop_diff.pop(self.SECURITY_SERVICES)
            s_curr = set([self.client_plugin().get_security_service(s).id
                          for s in self.properties.get(
                         self.SECURITY_SERVICES)])
            s_new = set([self.client_plugin().get_security_service(s).id
                        for s in services])
            for service in s_curr - s_new:
                self.client().share_networks.remove_security_service(
                    self.resource_id, service)
            for service in s_new - s_curr:
                self.client().share_networks.add_security_service(
                    self.resource_id, service)

        if prop_diff:
            neutron_subnet_id = prop_diff.get(self.NEUTRON_SUBNET)
            neutron_net_id = prop_diff.get(self.NEUTRON_NETWORK)
            if neutron_subnet_id and not neutron_net_id:
                neutron_net_id = self.client_plugin(
                    'neutron').network_id_from_subnet_id(neutron_subnet_id)

            self.client().share_networks.update(
                self.resource_id,
                name=prop_diff.get(self.NAME),
                neutron_net_id=neutron_net_id,
                neutron_subnet_id=neutron_subnet_id,
                description=prop_diff.get(self.DESCRIPTION))

    def parse_live_resource_data(self, resource_properties, resource_data):
        result = super(ManilaShareNetwork, self).parse_live_resource_data(
            resource_properties, resource_data)
        sec_list = self.client().security_services.list(
            search_opts={'share_network_id': self.resource_id})
        result.update({
            self.NOVA_NETWORK: resource_data.get('nova_net_id'),
            self.NEUTRON_NETWORK: resource_data.get('neutron_net_id'),
            self.NEUTRON_SUBNET: resource_data.get('neutron_subnet_id'),
            self.SECURITY_SERVICES: [service.id for service in sec_list]}
        )
        return result


def resource_mapping():
    return {'OS::Manila::ShareNetwork': ManilaShareNetwork}



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine.resources.openstack.keystone import role_assignments
from heat.engine import support
from heat.engine import translation


class KeystoneUser(resource.Resource,
                   role_assignments.KeystoneRoleAssignmentMixin):
    """Heat Template Resource for Keystone User.

    Users represent an individual API consumer. A user itself must be owned by
    a specific domain, and hence all user names are not globally unique, but
    only unique to their domain.
    """

    support_status = support.SupportStatus(
        version='2015.1',
        message=_('Supported versions: keystone v3'))

    default_client_name = 'keystone'

    entity = 'users'

    PROPERTIES = (
        NAME, DOMAIN, DESCRIPTION, ENABLED, EMAIL, PASSWORD,
        DEFAULT_PROJECT, GROUPS
    ) = (
        'name', 'domain', 'description', 'enabled', 'email', 'password',
        'default_project', 'groups'
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of keystone user.'),
            update_allowed=True
        ),
        DOMAIN: properties.Schema(
            properties.Schema.STRING,
            _('Name or ID of keystone domain.'),
            default='default',
            update_allowed=True,
            constraints=[constraints.CustomConstraint('keystone.domain')]
        ),
        DESCRIPTION: properties.Schema(
            properties.Schema.STRING,
            _('Description of keystone user.'),
            default='',
            update_allowed=True
        ),
        ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('Keystone user is enabled or disabled.'),
            default=True,
            update_allowed=True
        ),
        EMAIL: properties.Schema(
            properties.Schema.STRING,
            _('Email address of keystone user.'),
            update_allowed=True
        ),
        PASSWORD: properties.Schema(
            properties.Schema.STRING,
            _('Password of keystone user.'),
            update_allowed=True
        ),
        DEFAULT_PROJECT: properties.Schema(
            properties.Schema.STRING,
            _('Name or ID of default project of keystone user.'),
            update_allowed=True,
            constraints=[constraints.CustomConstraint('keystone.project')]
        ),
        GROUPS: properties.Schema(
            properties.Schema.LIST,
            _('Keystone user groups.'),
            update_allowed=True,
            schema=properties.Schema(
                properties.Schema.STRING,
                _('Keystone user group.'),
                constraints=[constraints.CustomConstraint('keystone.group')]
            )
        )
    }

    properties_schema.update(
        role_assignments.KeystoneRoleAssignmentMixin.mixin_properties_schema)

    ATTRIBUTES = (
        NAME_ATTR, DEFAULT_PROJECT_ATTR, DOMAIN_ATTR, ENABLED_ATTR,
        PASSWORD_EXPIRES_AT_ATTR
    ) = (
        'name', 'default_project_id', 'domain_id', 'enabled',
        'password_expires_at'
    )
    attributes_schema = {
        NAME_ATTR: attributes.Schema(
            _('User name.'),
            support_status=support.SupportStatus(version='9.0.0'),
            type=attributes.Schema.STRING
        ),
        DEFAULT_PROJECT_ATTR: attributes.Schema(
            _('Default project id for user.'),
            support_status=support.SupportStatus(version='9.0.0'),
            type=attributes.Schema.STRING
        ),
        DOMAIN_ATTR: attributes.Schema(
            _('Domain id for user.'),
            support_status=support.SupportStatus(version='9.0.0'),
            type=attributes.Schema.STRING
        ),
        ENABLED_ATTR: attributes.Schema(
            _('Flag of enable user.'),
            support_status=support.SupportStatus(version='9.0.0'),
            type=attributes.Schema.BOOLEAN
        ),
        PASSWORD_EXPIRES_AT_ATTR: attributes.Schema(
            _('Show user password expiration time.'),
            support_status=support.SupportStatus(version='9.0.0'),
            type=attributes.Schema.STRING
        ),
    }

    def translation_rules(self, properties):
        return [
            translation.TranslationRule(
                properties,
                translation.TranslationRule.RESOLVE,
                [self.DOMAIN],
                client_plugin=self.client_plugin(),
                finder='get_domain_id'
            ),
            translation.TranslationRule(
                properties,
                translation.TranslationRule.RESOLVE,
                [self.DEFAULT_PROJECT],
                client_plugin=self.client_plugin(),
                finder='get_project_id'
            ),
            translation.TranslationRule(
                properties,
                translation.TranslationRule.RESOLVE,
                [self.GROUPS],
                client_plugin=self.client_plugin(),
                finder='get_group_id'
            ),
        ]

    def validate(self):
        super(KeystoneUser, self).validate()
        self.validate_assignment_properties()

    def client(self):
        return super(KeystoneUser, self).client().client

    def _update_user(self,
                     user_id,
                     domain,
                     new_name=None,
                     new_description=None,
                     new_email=None,
                     new_password=None,
                     new_default_project=None,
                     enabled=None):
        values = dict()

        if new_name is not None:
            values['name'] = new_name
        if new_description is not None:
            values['description'] = new_description
        if new_email is not None:
            values['email'] = new_email
        if new_password is not None:
            values['password'] = new_password
        if new_default_project is not None:
            values['default_project'] = new_default_project
        if enabled is not None:
            values['enabled'] = enabled

        # If there're no args above, keystone raises BadRequest error with
        # message about not enough parameters for updating, so return from
        # this method to prevent raising error.
        if not values:
            return

        values['user'] = user_id
        values['domain'] = domain

        return self.client().users.update(**values)

    def _add_user_to_groups(self, user_id, groups):
        if groups is not None:
            for group_id in groups:
                self.client().users.add_to_group(user_id,
                                                 group_id)

    def _remove_user_from_groups(self, user_id, groups):
        if groups is not None:
            for group_id in groups:
                self.client().users.remove_from_group(user_id,
                                                      group_id)

    def _find_diff(self, updated_prps, stored_prps):
        new_group_ids = list(set(updated_prps or []) - set(stored_prps or []))

        removed_group_ids = list(set(stored_prps or []) -
                                 set(updated_prps or []))

        return new_group_ids, removed_group_ids

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        user = self.client().users.get(self.resource_id)
        return getattr(user, name, None)

    def handle_create(self):
        user_name = (self.properties[self.NAME] or
                     self.physical_resource_name())
        description = self.properties[self.DESCRIPTION]
        domain = self.properties[self.DOMAIN]
        enabled = self.properties[self.ENABLED]
        email = self.properties[self.EMAIL]
        password = self.properties[self.PASSWORD]
        default_project = self.properties[self.DEFAULT_PROJECT]
        groups = self.properties[self.GROUPS]

        user = self.client().users.create(
            name=user_name,
            domain=domain,
            description=description,
            enabled=enabled,
            email=email,
            password=password,
            default_project=default_project)

        self.resource_id_set(user.id)

        self._add_user_to_groups(user.id, groups)

        self.create_assignment(user_id=user.id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            name = None
            # Don't update the name if no change
            if self.NAME in prop_diff:
                name = prop_diff[self.NAME] or self.physical_resource_name()

            description = prop_diff.get(self.DESCRIPTION)
            enabled = prop_diff.get(self.ENABLED)
            email = prop_diff.get(self.EMAIL)
            password = prop_diff.get(self.PASSWORD)
            domain = (prop_diff.get(self.DOMAIN)
                      or self.properties[self.DOMAIN])

            default_project = prop_diff.get(self.DEFAULT_PROJECT)

            self._update_user(
                user_id=self.resource_id,
                domain=domain,
                new_name=name,
                new_description=description,
                enabled=enabled,
                new_default_project=default_project,
                new_email=email,
                new_password=password
            )

            if self.GROUPS in prop_diff:
                (new_group_ids, removed_group_ids) = self._find_diff(
                    prop_diff[self.GROUPS],
                    self.properties[self.GROUPS])
                if new_group_ids:
                    self._add_user_to_groups(self.resource_id, new_group_ids)

                if removed_group_ids:
                    self._remove_user_from_groups(self.resource_id,
                                                  removed_group_ids)

            self.update_assignment(prop_diff=prop_diff,
                                   user_id=self.resource_id)

    def parse_live_resource_data(self, resource_properties, resource_data):
        user_reality = {
            self.ROLES: self.parse_list_assignments(user_id=self.resource_id),
            self.DEFAULT_PROJECT: resource_data.get('default_project_id'),
            self.DOMAIN: resource_data.get('domain_id'),
            self.GROUPS: [group.id for group in self.client().groups.list(
                user=self.resource_id)]
        }
        props_keys = [self.NAME, self.DESCRIPTION, self.ENABLED, self.EMAIL]
        for key in props_keys:
            user_reality.update({key: resource_data.get(key)})
        return user_reality

    def handle_delete(self):
        if self.resource_id is not None:
            with self.client_plugin().ignore_not_found:
                if self.properties[self.GROUPS] is not None:
                    self._remove_user_from_groups(
                        self.resource_id,
                        self.properties[self.GROUPS])

                self.client().users.delete(self.resource_id)


def resource_mapping():
    return {
        'OS::Keystone::User': KeystoneUser
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from heat.common import identifier
from heat.tests import common


class IdentifierTest(common.HeatTestCase):
    url_prefix = 'http://1.2.3.4/foo/'

    def test_attrs(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p')
        self.assertEqual('t', hi.tenant)
        self.assertEqual('s', hi.stack_name)
        self.assertEqual('i', hi.stack_id)
        self.assertEqual('/p', hi.path)

    def test_path_default(self):
        hi = identifier.HeatIdentifier('t', 's', 'i')
        self.assertEqual('', hi.path)

    def test_items(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p')
        self.assertEqual('t', hi['tenant'])
        self.assertEqual('s', hi['stack_name'])
        self.assertEqual('i', hi['stack_id'])
        self.assertEqual('/p', hi['path'])

    def test_invalid_attr(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p')
        hi.identity['foo'] = 'bar'
        self.assertRaises(AttributeError, getattr, hi, 'foo')

    def test_invalid_item(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p')
        hi.identity['foo'] = 'bar'
        self.assertRaises(KeyError, lambda o, k: o[k], hi, 'foo')

    def test_stack_path(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p')
        self.assertEqual('s/i', hi.stack_path())

    def test_arn(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p')
        self.assertEqual('arn:openstack:heat::t:stacks/s/i/p', hi.arn())

    def test_arn_url(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p')
        self.assertEqual('/arn%3Aopenstack%3Aheat%3A%3At%3Astacks/s/i/p',
                         hi.arn_url_path())

    def test_arn_id_int(self):
        hi = identifier.HeatIdentifier('t', 's', 42, 'p')
        self.assertEqual('arn:openstack:heat::t:stacks/s/42/p', hi.arn())

    def test_arn_parse(self):
        arn = 'arn:openstack:heat::t:stacks/s/i/p'
        hi = identifier.HeatIdentifier.from_arn(arn)
        self.assertEqual('t', hi.tenant)
        self.assertEqual('s', hi.stack_name)
        self.assertEqual('i', hi.stack_id)
        self.assertEqual('/p', hi.path)

    def test_arn_url_parse(self):
        url = self.url_prefix + 'arn%3Aopenstack%3Aheat%3A%3At%3Astacks/s/i/p'
        hi = identifier.HeatIdentifier.from_arn_url(url)
        self.assertEqual('t', hi.tenant)
        self.assertEqual('s', hi.stack_name)
        self.assertEqual('i', hi.stack_id)
        self.assertEqual('/p', hi.path)

    def test_arn_parse_path_default(self):
        arn = 'arn:openstack:heat::t:stacks/s/i'
        hi = identifier.HeatIdentifier.from_arn(arn)
        self.assertEqual('t', hi.tenant)
        self.assertEqual('s', hi.stack_name)
        self.assertEqual('i', hi.stack_id)
        self.assertEqual('', hi.path)

    def test_arn_url_parse_default(self):
        url = self.url_prefix + 'arn%3Aopenstack%3Aheat%3A%3At%3Astacks/s/i'
        hi = identifier.HeatIdentifier.from_arn_url(url)
        self.assertEqual('t', hi.tenant)
        self.assertEqual('s', hi.stack_name)
        self.assertEqual('i', hi.stack_id)
        self.assertEqual('', hi.path)

    def test_arn_parse_upper(self):
        arn = 'ARN:openstack:heat::t:stacks/s/i/p'
        hi = identifier.HeatIdentifier.from_arn(arn)
        self.assertEqual('s', hi.stack_name)
        self.assertEqual('i', hi.stack_id)
        self.assertEqual('/p', hi.path)

    def test_arn_url_parse_upper(self):
        url = self.url_prefix + 'ARN%3Aopenstack%3Aheat%3A%3At%3Astacks/s/i/p'
        hi = identifier.HeatIdentifier.from_arn_url(url)
        self.assertEqual('t', hi.tenant)
        self.assertEqual('s', hi.stack_name)
        self.assertEqual('i', hi.stack_id)
        self.assertEqual('/p', hi.path)

    def test_arn_url_parse_qs(self):
        url = (self.url_prefix +
               'arn%3Aopenstack%3Aheat%3A%3At%3Astacks/s/i/p?foo=bar')
        hi = identifier.HeatIdentifier.from_arn_url(url)
        self.assertEqual('t', hi.tenant)
        self.assertEqual('s', hi.stack_name)
        self.assertEqual('i', hi.stack_id)
        self.assertEqual('/p', hi.path)

    def test_arn_parse_arn_invalid(self):
        arn = 'urn:openstack:heat::t:stacks/s/i'
        self.assertRaises(ValueError, identifier.HeatIdentifier.from_arn, arn)

    def test_arn_url_parse_arn_invalid(self):
        url = self.url_prefix + 'urn:openstack:heat::t:stacks/s/i/p'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_parse_os_invalid(self):
        arn = 'arn:aws:heat::t:stacks/s/i'
        self.assertRaises(ValueError, identifier.HeatIdentifier.from_arn, arn)

    def test_arn_url_parse_os_invalid(self):
        url = self.url_prefix + 'arn:aws:heat::t:stacks/s/i/p'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_parse_heat_invalid(self):
        arn = 'arn:openstack:cool::t:stacks/s/i'
        self.assertRaises(ValueError, identifier.HeatIdentifier.from_arn, arn)

    def test_arn_url_parse_heat_invalid(self):
        url = self.url_prefix + 'arn:openstack:cool::t:stacks/s/i/p'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_parse_stacks_invalid(self):
        arn = 'arn:openstack:heat::t:sticks/s/i'
        self.assertRaises(ValueError, identifier.HeatIdentifier.from_arn, arn)

    def test_arn_url_parse_stacks_invalid(self):
        url = self.url_prefix + 'arn%3Aopenstack%3Aheat%3A%3At%3Asticks/s/i/p'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_parse_missing_field(self):
        arn = 'arn:openstack:heat::t:stacks/s'
        self.assertRaises(ValueError, identifier.HeatIdentifier.from_arn, arn)

    def test_arn_url_parse_missing_field(self):
        url = self.url_prefix + 'arn%3Aopenstack%3Aheat%3A%3At%3Asticks/s/'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_parse_empty_field(self):
        arn = 'arn:openstack:heat::t:stacks//i'
        self.assertRaises(ValueError, identifier.HeatIdentifier.from_arn, arn)

    def test_arn_url_parse_empty_field(self):
        url = self.url_prefix + 'arn%3Aopenstack%3Aheat%3A%3At%3Asticks//i'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_url_parse_leading_char(self):
        url = self.url_prefix + 'Aarn%3Aopenstack%3Aheat%3A%3At%3Asticks/s/i/p'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_url_parse_leading_space(self):
        url = self.url_prefix + ' arn%3Aopenstack%3Aheat%3A%3At%3Asticks/s/i/p'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_url_parse_badurl_proto(self):
        url = 'htt://1.2.3.4/foo/arn%3Aopenstack%3Aheat%3A%3At%3Asticks/s/i/p'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_url_parse_badurl_host(self):
        url = 'http:///foo/arn%3Aopenstack%3Aheat%3A%3At%3Asticks/s/i/p'
        self.assertRaises(ValueError,
                          identifier.HeatIdentifier.from_arn_url, url)

    def test_arn_round_trip(self):
        hii = identifier.HeatIdentifier('t', 's', 'i', 'p')
        hio = identifier.HeatIdentifier.from_arn(hii.arn())
        self.assertEqual(hii.tenant, hio.tenant)
        self.assertEqual(hii.stack_name, hio.stack_name)
        self.assertEqual(hii.stack_id, hio.stack_id)
        self.assertEqual(hii.path, hio.path)

    def test_arn_parse_round_trip(self):
        arn = 'arn:openstack:heat::t:stacks/s/i/p'
        hi = identifier.HeatIdentifier.from_arn(arn)
        self.assertEqual(arn, hi.arn())

    def test_arn_url_parse_round_trip(self):
        arn = '/arn%3Aopenstack%3Aheat%3A%3At%3Astacks/s/i/p'
        url = 'http://1.2.3.4/foo' + arn
        hi = identifier.HeatIdentifier.from_arn_url(url)
        self.assertEqual(arn, hi.arn_url_path())

    def test_dict_round_trip(self):
        hii = identifier.HeatIdentifier('t', 's', 'i', 'p')
        hio = identifier.HeatIdentifier(**dict(hii))
        self.assertEqual(hii.tenant, hio.tenant)
        self.assertEqual(hii.stack_name, hio.stack_name)
        self.assertEqual(hii.stack_id, hio.stack_id)
        self.assertEqual(hii.path, hio.path)

    def test_url_path(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p')
        self.assertEqual('t/stacks/s/i/p', hi.url_path())

    def test_url_path_default(self):
        hi = identifier.HeatIdentifier('t', 's', 'i')
        self.assertEqual('t/stacks/s/i', hi.url_path())

    def test_url_path_with_unicode_path(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', '\u5de5')
        self.assertEqual('t/stacks/s/i/%E5%B7%A5', hi.url_path())

    def test_tenant_escape(self):
        hi = identifier.HeatIdentifier(':/', 's', 'i')
        self.assertEqual(':/', hi.tenant)
        self.assertEqual('%3A%2F/stacks/s/i', hi.url_path())
        self.assertEqual('arn:openstack:heat::%3A%2F:stacks/s/i', hi.arn())

    def test_name_escape(self):
        hi = identifier.HeatIdentifier('t', ':%', 'i')
        self.assertEqual(':%', hi.stack_name)
        self.assertEqual('t/stacks/%3A%25/i', hi.url_path())
        self.assertEqual('arn:openstack:heat::t:stacks/%3A%25/i', hi.arn())

    def test_id_escape(self):
        hi = identifier.HeatIdentifier('t', 's', ':/')
        self.assertEqual(':/', hi.stack_id)
        self.assertEqual('t/stacks/s/%3A%2F', hi.url_path())
        self.assertEqual('arn:openstack:heat::t:stacks/s/%3A%2F', hi.arn())

    def test_id_contains(self):
        hi = identifier.HeatIdentifier('t', 's', ':/')
        self.assertNotIn("t", hi)
        self.assertIn("stack_id", hi)

    def test_path_escape(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', ':/')
        self.assertEqual('/:/', hi.path)
        self.assertEqual('t/stacks/s/i/%3A/', hi.url_path())
        self.assertEqual('arn:openstack:heat::t:stacks/s/i/%3A/', hi.arn())

    def test_tenant_decode(self):
        arn = 'arn:openstack:heat::%3A%2F:stacks/s/i'
        hi = identifier.HeatIdentifier.from_arn(arn)
        self.assertEqual(':/', hi.tenant)

    def test_url_tenant_decode(self):
        enc_arn = 'arn%3Aopenstack%3Aheat%3A%3A%253A%252F%3Astacks%2Fs%2Fi'
        url = self.url_prefix + enc_arn
        hi = identifier.HeatIdentifier.from_arn_url(url)
        self.assertEqual(':/', hi.tenant)

    def test_name_decode(self):
        arn = 'arn:openstack:heat::t:stacks/%3A%25/i'
        hi = identifier.HeatIdentifier.from_arn(arn)
        self.assertEqual(':%', hi.stack_name)

    def test_url_name_decode(self):
        enc_arn = 'arn%3Aopenstack%3Aheat%3A%3At%3Astacks%2F%253A%2525%2Fi'
        url = self.url_prefix + enc_arn
        hi = identifier.HeatIdentifier.from_arn_url(url)
        self.assertEqual(':%', hi.stack_name)

    def test_id_decode(self):
        arn = 'arn:openstack:heat::t:stacks/s/%3A%2F'
        hi = identifier.HeatIdentifier.from_arn(arn)
        self.assertEqual(':/', hi.stack_id)

    def test_url_id_decode(self):
        enc_arn = 'arn%3Aopenstack%3Aheat%3A%3At%3Astacks%2Fs%2F%253A%252F'
        url = self.url_prefix + enc_arn
        hi = identifier.HeatIdentifier.from_arn_url(url)
        self.assertEqual(':/', hi.stack_id)

    def test_path_decode(self):
        arn = 'arn:openstack:heat::t:stacks/s/i/%3A%2F'
        hi = identifier.HeatIdentifier.from_arn(arn)
        self.assertEqual('/:/', hi.path)

    def test_url_path_decode(self):
        enc_arn = 'arn%3Aopenstack%3Aheat%3A%3At%3Astacks%2Fs%2Fi%2F%253A%252F'
        url = self.url_prefix + enc_arn
        hi = identifier.HeatIdentifier.from_arn_url(url)
        self.assertEqual('/:/', hi.path)

    def test_arn_escape_decode_round_trip(self):
        hii = identifier.HeatIdentifier(':/', ':%', ':/', ':/')
        hio = identifier.HeatIdentifier.from_arn(hii.arn())
        self.assertEqual(hii.tenant, hio.tenant)
        self.assertEqual(hii.stack_name, hio.stack_name)
        self.assertEqual(hii.stack_id, hio.stack_id)
        self.assertEqual(hii.path, hio.path)

    def test_arn_decode_escape_round_trip(self):
        arn = 'arn:openstack:heat::%3A%2F:stacks/%3A%25/%3A%2F/%3A/'
        hi = identifier.HeatIdentifier.from_arn(arn)
        self.assertEqual(arn, hi.arn())

    def test_arn_url_decode_escape_round_trip(self):
        enc_arn = "".join(['arn%3Aopenstack%3Aheat%3A%3A%253A%252F%3A',
                           'stacks%2F%253A%2525%2F%253A%252F%2F%253A'])
        url = self.url_prefix + enc_arn
        hi = identifier.HeatIdentifier.from_arn_url(url)
        hi2 = identifier.HeatIdentifier.from_arn_url(self.url_prefix +
                                                     hi.arn_url_path())
        self.assertEqual(hi, hi2)

    def test_stack_name_slash(self):
        self.assertRaises(ValueError, identifier.HeatIdentifier,
                          't', 's/s', 'i', 'p')

    def test_equal(self):
        hi1 = identifier.HeatIdentifier('t', 's', 'i', 'p')
        hi2 = identifier.HeatIdentifier('t', 's', 'i', 'p')
        self.assertTrue(hi1 == hi2)

    def test_equal_dict(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p')
        self.assertTrue(hi == dict(hi))
        self.assertTrue(dict(hi) == hi)

    def test_not_equal(self):
        hi1 = identifier.HeatIdentifier('t', 's', 'i', 'p')
        hi2 = identifier.HeatIdentifier('t', 's', 'i', 'q')
        self.assertFalse(hi1 == hi2)
        self.assertFalse(hi2 == hi1)

    def test_not_equal_dict(self):
        hi1 = identifier.HeatIdentifier('t', 's', 'i', 'p')
        hi2 = identifier.HeatIdentifier('t', 's', 'i', 'q')
        self.assertFalse(hi1 == dict(hi2))
        self.assertFalse(dict(hi1) == hi2)
        self.assertFalse(hi1 == {'tenant': 't',
                                 'stack_name': 's',
                                 'stack_id': 'i'})
        self.assertFalse({'tenant': 't',
                          'stack_name': 's',
                          'stack_id': 'i'} == hi1)

    def test_path_components(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', 'p1/p2/p3')
        self.assertEqual(['p1', 'p2', 'p3'], hi._path_components())


class ResourceIdentifierTest(common.HeatTestCase):
    def test_resource_init_no_path(self):
        si = identifier.HeatIdentifier('t', 's', 'i')
        ri = identifier.ResourceIdentifier(resource_name='r', **si)
        self.assertEqual('/resources/r', ri.path)

    def test_resource_init_path(self):
        si = identifier.HeatIdentifier('t', 's', 'i')
        pi = identifier.ResourceIdentifier(resource_name='p', **si)
        ri = identifier.ResourceIdentifier(resource_name='r', **pi)
        self.assertEqual('/resources/p/resources/r', ri.path)

    def test_resource_init_from_dict(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', '/resources/r')
        ri = identifier.ResourceIdentifier(**hi)
        self.assertEqual(hi, ri)

    def test_resource_stack(self):
        si = identifier.HeatIdentifier('t', 's', 'i')
        ri = identifier.ResourceIdentifier(resource_name='r', **si)
        self.assertEqual(si, ri.stack())

    def test_resource_id(self):
        ri = identifier.ResourceIdentifier('t', 's', 'i', '', 'r')
        self.assertEqual('r', ri.resource_name)

    def test_resource_name_slash(self):
        self.assertRaises(ValueError, identifier.ResourceIdentifier,
                          't', 's', 'i', 'p', 'r/r')


class EventIdentifierTest(common.HeatTestCase):
    def test_event_init_integer_id(self):
        self._test_event_init('42')

    def test_event_init_uuid_id(self):
        self._test_event_init('a3455d8c-9f88-404d-a85b-5315293e67de')

    def _test_event_init(self, event_id):
        si = identifier.HeatIdentifier('t', 's', 'i')
        pi = identifier.ResourceIdentifier(resource_name='p', **si)
        ei = identifier.EventIdentifier(event_id=event_id, **pi)
        self.assertEqual('/resources/p/events/{0}'.format(event_id), ei.path)

    def test_event_init_from_dict(self):
        hi = identifier.HeatIdentifier('t', 's', 'i', '/resources/p/events/42')
        ei = identifier.EventIdentifier(**hi)
        self.assertEqual(hi, ei)

    def test_event_stack(self):
        si = identifier.HeatIdentifier('t', 's', 'i')
        pi = identifier.ResourceIdentifier(resource_name='r', **si)
        ei = identifier.EventIdentifier(event_id='e', **pi)
        self.assertEqual(si, ei.stack())

    def test_event_resource(self):
        si = identifier.HeatIdentifier('t', 's', 'i')
        pi = identifier.ResourceIdentifier(resource_name='r', **si)
        ei = identifier.EventIdentifier(event_id='e', **pi)
        self.assertEqual(pi, ei.resource())

    def test_resource_name(self):
        ei = identifier.EventIdentifier('t', 's', 'i', '/resources/p', 'e')
        self.assertEqual('p', ei.resource_name)

    def test_event_id_integer(self):
        self._test_event_id('42')

    def test_event_id_uuid(self):
        self._test_event_id('a3455d8c-9f88-404d-a85b-5315293e67de')

    def _test_event_id(self, event_id):
        ei = identifier.EventIdentifier('t', 's', 'i', '/resources/p',
                                        event_id)
        self.assertEqual(event_id, ei.event_id)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""A fake FakeKeystoneClient. This can be used during some runtime
scenarios where you want to disable Heat's internal Keystone dependencies
entirely. One example is the TripleO Undercloud installer.

To use this class at runtime set to following heat.conf config setting:

  keystone_backend = heat.engine.clients.os.keystone.fake_keystoneclient\
  .FakeKeystoneClient

"""

from keystoneauth1 import session

from heat.common import context


class FakeKeystoneClient(object):
    def __init__(self, username='test_username', password='password',
                 user_id='1234', access='4567', secret='8901',
                 credential_id='abcdxyz', auth_token='abcd1234',
                 context=None, stack_domain_id='4321', client=None):
        self.username = username
        self.password = password
        self.user_id = user_id
        self.access = access
        self.secret = secret
        self.session = session.Session()
        self.credential_id = credential_id
        self.token = auth_token
        self.context = context
        self.v3_endpoint = 'http://localhost:5000/v3'
        self.stack_domain_id = stack_domain_id
        self.client = client

        class FakeCred(object):
            id = self.credential_id
            access = self.access
            secret = self.secret
        self.creds = FakeCred()

    def create_stack_user(self, username, password):
        self.username = username
        return self.user_id

    def delete_stack_user(self, user_id):
        self.user_id = None

    def get_ec2_keypair(self, access, user_id):
        if user_id == self.user_id:
            if access == self.access:
                return self.creds
            else:
                raise ValueError("Unexpected access %s" % access)
        else:
            raise ValueError("Unexpected user_id %s" % user_id)

    def create_ec2_keypair(self, user_id):
        if user_id == self.user_id:
            return self.creds

    def delete_ec2_keypair(self, credential_id=None, user_id=None,
                           access=None):
        if user_id == self.user_id and access == self.creds.access:
            self.creds = None
        else:
            raise Exception('Incorrect user_id or access')

    def enable_stack_user(self, user_id):
        pass

    def disable_stack_user(self, user_id):
        pass

    def create_trust_context(self):
        return context.RequestContext(username=self.username,
                                      password=self.password,
                                      is_admin=False,
                                      trust_id='atrust',
                                      trustor_user_id=self.user_id)

    def regenerate_trust_context(self):
        return context.RequestContext(username=self.username,
                                      password=self.password,
                                      is_admin=False,
                                      trust_id='atrust',
                                      trustor_user_id=self.user_id)

    def delete_trust(self, trust_id):
        pass

    def delete_stack_domain_project(self, project_id):
        pass

    def create_stack_domain_project(self, stack_id):
        return 'aprojectid'

    def create_stack_domain_user(self, username, project_id, password=None):
        return self.user_id

    def delete_stack_domain_user(self, user_id, project_id):
        pass

    def create_stack_domain_user_keypair(self, user_id, project_id):
        return self.creds

    def enable_stack_domain_user(self, user_id, project_id):
        pass

    def disable_stack_domain_user(self, user_id, project_id):
        pass

    def delete_stack_domain_user_keypair(self, user_id, project_id,
                                         credential_id):
        pass

    def stack_domain_user_token(self, user_id, project_id, password):
        return 'adomainusertoken'

    def server_keystone_endpoint_url(self, fallback_endpoint):
        return fallback_endpoint



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
from unittest import mock


from heat.common import template_format
from heat.engine.resources.openstack.manila import share_type as mshare_type
from heat.engine import rsrc_defn
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils

manila_template = """
heat_template_version: 2013-05-23
resources:
  test_share_type:
    type: OS::Manila::ShareType
    properties:
      name: test_share_type
      driver_handles_share_servers: True
      extra_specs: {"test":"test"}
      is_public: False
      snapshot_support: True
"""


class DummyShare(object):
    def __init__(self):
        self.to_dict = lambda: {'attr': 'val'}


class ManilaShareTypeTest(common.HeatTestCase):

    def _init_share(self, stack_name, share_type_name="test_share_type"):
        # parse stack
        tmp = template_format.parse(manila_template)
        self.stack = utils.parse_stack(tmp, stack_name=stack_name)
        defns = self.stack.t.resource_definitions(self.stack)
        res_def = defns["test_share_type"]
        share_type = mshare_type.ManilaShareType(
            share_type_name, res_def, self.stack)
        # mock clients and plugins
        mock_client = mock.MagicMock()
        client = mock.MagicMock(return_value=mock_client)
        share_type.client = client

        return share_type

    def test_share_type_create(self):
        share_type = self._init_share("stack_share_type_create")
        fake_share_type = mock.MagicMock(id="type_id")
        share_type.client().share_types.create.return_value = fake_share_type
        scheduler.TaskRunner(share_type.create)()
        self.assertEqual("type_id", share_type.resource_id)
        share_type.client().share_types.create.assert_called_once_with(
            name="test_share_type", spec_driver_handles_share_servers=True,
            is_public=False, spec_snapshot_support=True)
        fake_share_type.set_keys.assert_called_once_with({"test": "test"})
        self.assertEqual('share_types', share_type.entity)

    def test_share_type_update(self):
        share_type = self._init_share("stack_share_type_update")
        share_type.client().share_types.create.return_value = mock.MagicMock(
            id="type_id")
        fake_share_type = mock.MagicMock()
        share_type.client().share_types.get.return_value = fake_share_type
        scheduler.TaskRunner(share_type.create)()
        updated_props = copy.deepcopy(share_type.properties.data)
        updated_props[mshare_type.ManilaShareType.EXTRA_SPECS] = {
            "fake_key": "fake_value"}
        after = rsrc_defn.ResourceDefinition(share_type.name,
                                             share_type.type(), updated_props)
        scheduler.TaskRunner(share_type.update, after)()
        fake_share_type.unset_keys.assert_called_once_with({"test": "test"})
        fake_share_type.set_keys.assert_called_with(
            updated_props[mshare_type.ManilaShareType.EXTRA_SPECS])

    def test_get_live_state(self):
        share_type = self._init_share("stack_share_type_update")

        value = mock.MagicMock()
        value.to_dict.return_value = {
            'os-share-type-access:is_public': True,
            'required_extra_specs': {},
            'extra_specs': {'test': 'test',
                            'snapshot_support': 'True',
                            'driver_handles_share_servers': 'True'},
            'id': 'cc76cb22-75fe-4e6e-b618-7c345b2444e3',
            'name': 'test'}

        share_type.client().share_types.get.return_value = value

        reality = share_type.get_live_state(share_type.properties)
        expected = {
            'extra_specs': {'test': 'test'}
        }

        self.assertEqual(set(expected.keys()), set(reality.keys()))
        for key in expected:
            self.assertEqual(expected[key], reality[key])



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import template_format
from heat.engine import resource
from heat.engine.resources.openstack.heat import none_resource as none
from heat.engine import scheduler
from heat.tests import common
from heat.tests import utils


class NoneResourceTest(common.HeatTestCase):

    tmpl = '''
heat_template_version: 2015-10-15
resources:
  none:
    type: OS::Heat::None
    properties:
      ignored: foo
outputs:
  anything:
    value: {get_attr: [none, anything]}
'''

    def _create_none_stack(self):
        self.t = template_format.parse(self.tmpl)
        self.stack = utils.parse_stack(self.t)
        self.rsrc = self.stack['none']
        self.assertIsNone(self.rsrc.validate())
        self.stack.create()
        self.assertEqual(self.rsrc.CREATE, self.rsrc.action)
        self.assertEqual(self.rsrc.COMPLETE, self.rsrc.status)
        self.assertEqual(self.stack.CREATE, self.stack.action)
        self.assertEqual(self.stack.COMPLETE, self.stack.status)
        self.stack._update_all_resource_data(False, True)
        self.assertIsNone(self.stack.outputs['anything'].get_value())

    def test_none_stack_create(self):
        self._create_none_stack()

    def test_none_stack_update_nochange(self):
        self._create_none_stack()
        before_refid = self.rsrc.FnGetRefId()
        self.assertIsNotNone(before_refid)
        utils.update_stack(self.stack, self.t)
        self.assertEqual((self.stack.UPDATE, self.stack.COMPLETE),
                         self.stack.state)
        self.assertEqual(before_refid, self.stack['none'].FnGetRefId())

    def test_none_stack_update_add_prop(self):
        self._create_none_stack()
        before_refid = self.rsrc.FnGetRefId()
        self.assertIsNotNone(before_refid)
        new_t = self.t.copy()
        new_t['resources']['none']['properties']['another'] = 123
        utils.update_stack(self.stack, new_t)
        self.assertEqual((self.stack.UPDATE, self.stack.COMPLETE),
                         self.stack.state)
        self.assertEqual(before_refid, self.stack['none'].FnGetRefId())

    def test_none_stack_update_del_prop(self):
        self._create_none_stack()
        before_refid = self.rsrc.FnGetRefId()
        self.assertIsNotNone(before_refid)
        new_t = self.t.copy()
        del new_t['resources']['none']['properties']['ignored']
        utils.update_stack(self.stack, new_t)
        self.assertEqual((self.stack.UPDATE, self.stack.COMPLETE),
                         self.stack.state)
        self.assertEqual(before_refid, self.stack['none'].FnGetRefId())


class PlaceholderResourceTest(common.HeatTestCase):

    tmpl = '''
heat_template_version: 2015-10-15
resources:
  none:
    type: OS::BAR::FOO
    properties:
      ignored: foo
'''

    class FooResource(none.NoneResource):
        default_client_name = 'heat'
        entity = 'foo'

    FOO_RESOURCE_TYPE = 'OS::BAR::FOO'

    def setUp(self):
        super(PlaceholderResourceTest, self).setUp()
        resource._register_class(self.FOO_RESOURCE_TYPE, self.FooResource)
        self.t = template_format.parse(self.tmpl)
        self.stack = utils.parse_stack(self.t)
        self.rsrc = self.stack['none']
        self.client = mock.MagicMock()
        self.patchobject(self.FooResource, 'client', return_value=self.client)
        scheduler.TaskRunner(self.rsrc.create)()

    def _test_delete(self, is_placeholder=True):
        if not is_placeholder:
            delete_call_count = 1
            self.rsrc.data = mock.Mock(
                return_value={})
        else:
            delete_call_count = 0
            self.rsrc.data = mock.Mock(
                return_value={'is_placeholder': 'True'})
        scheduler.TaskRunner(self.rsrc.delete)()
        self.assertEqual((self.rsrc.DELETE, self.rsrc.COMPLETE),
                         self.rsrc.state)
        self.assertEqual(delete_call_count, self.client.foo.delete.call_count)
        self.assertEqual('foo', self.rsrc.entity)

    def test_not_placeholder_resource_delete(self):
        self._test_delete(is_placeholder=False)

    def test_placeholder_resource_delete(self):
        self._test_delete()



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from unittest import mock
import yaml

from osc_lib import exceptions

from heat.common import exception
from heat.common import template_format
from heat.engine.resources.openstack.octavia import listener
from heat.tests import common
from heat.tests.openstack.octavia import inline_templates
from heat.tests import utils


class ListenerTest(common.HeatTestCase):

    def test_resource_mapping(self):
        mapping = listener.resource_mapping()
        self.assertEqual(listener.Listener,
                         mapping['OS::Octavia::Listener'])

    def _create_stack(self, tmpl=inline_templates.LISTENER_TEMPLATE):
        self.t = template_format.parse(tmpl)
        self.stack = utils.parse_stack(self.t)
        self.listener = self.stack['listener']

        self.octavia_client = mock.MagicMock()
        self.listener.client = mock.MagicMock(return_value=self.octavia_client)
        self.listener.client_plugin().client = mock.MagicMock(
            return_value=self.octavia_client)

    def test_validate_terminated_https(self):
        tmpl = yaml.safe_load(inline_templates.LISTENER_TEMPLATE)
        props = tmpl['resources']['listener']['properties']
        props['protocol'] = 'TERMINATED_HTTPS'
        del props['default_tls_container_ref']
        self._create_stack(tmpl=yaml.safe_dump(tmpl))

        self.assertRaises(exception.StackValidationFailed,
                          self.listener.validate)

    def test_create(self):
        self._create_stack()
        self.octavia_client.listener_show.side_effect = [
            {'provisioning_status': 'PENDING_CREATE'},
            {'provisioning_status': 'PENDING_CREATE'},
            {'provisioning_status': 'ACTIVE'},
        ]
        self.octavia_client.listener_create.side_effect = [
            exceptions.Conflict(409), {'listener': {'id': '1234'}}
        ]
        expected = {
            'listener': {
                'protocol_port': 80,
                'protocol': 'TCP',
                'loadbalancer_id': '123',
                'default_pool_id': 'my_pool',
                'name': 'my_listener',
                'description': 'my listener',
                'admin_state_up': True,
                'default_tls_container_ref': 'ref',
                'sni_container_refs': ['ref1', 'ref2'],
                'connection_limit': -1,
                'tenant_id': '1234',
                'allowed_cidrs': ['10.10.0.0/16', '192.168.0.0/16']
            }
        }

        props = self.listener.handle_create()

        self.assertFalse(self.listener.check_create_complete(props))
        self.octavia_client.listener_create.assert_called_with(json=expected)
        self.assertFalse(self.listener.check_create_complete(props))
        self.octavia_client.listener_create.assert_called_with(json=expected)
        self.assertFalse(self.listener.check_create_complete(props))
        self.assertTrue(self.listener.check_create_complete(props))

    def test_create_missing_properties(self):
        for prop in ('protocol', 'protocol_port', 'loadbalancer'):
            tmpl = yaml.safe_load(inline_templates.LISTENER_TEMPLATE)
            del tmpl['resources']['listener']['properties'][prop]
            self._create_stack(tmpl=yaml.safe_dump(tmpl))
            self.assertRaises(exception.StackValidationFailed,
                              self.listener.validate)

    def test_show_resource(self):
        self._create_stack()
        self.listener.resource_id_set('1234')
        self.octavia_client.listener_show.return_value = {'id': '1234'}
        self.assertEqual({'id': '1234'}, self.listener._show_resource())

        self.octavia_client.listener_show.assert_called_with('1234')

    def test_update(self):
        self._create_stack()
        self.listener.resource_id_set('1234')
        self.octavia_client.listener_show.side_effect = [
            {'provisioning_status': 'PENDING_UPDATE'},
            {'provisioning_status': 'PENDING_UPDATE'},
            {'provisioning_status': 'ACTIVE'},
        ]
        self.octavia_client.listener_set.side_effect = [
            exceptions.Conflict(409), None]
        prop_diff = {
            'admin_state_up': False,
            'name': 'your_listener',
        }

        prop_diff = self.listener.handle_update(self.listener.t,
                                                None, prop_diff)

        self.assertFalse(self.listener.check_update_complete(prop_diff))
        self.assertFalse(self.listener._update_called)
        self.octavia_client.listener_set.assert_called_with(
            '1234', json={'listener': prop_diff})
        self.assertFalse(self.listener.check_update_complete(prop_diff))
        self.assertTrue(self.listener._update_called)
        self.octavia_client.listener_set.assert_called_with(
            '1234', json={'listener': prop_diff})
        self.assertFalse(self.listener.check_update_complete(prop_diff))
        self.assertTrue(self.listener.check_update_complete(prop_diff))

    def test_delete(self):
        self._create_stack()
        self.listener.resource_id_set('1234')
        self.octavia_client.listener_show.side_effect = [
            {'provisioning_status': 'PENDING_DELETE'},
            {'provisioning_status': 'PENDING_DELETE'},
            {'provisioning_status': 'DELETED'},
        ]
        self.octavia_client.listener_delete.side_effect = [
            exceptions.Conflict(409), None]

        self.listener.handle_delete()

        self.assertFalse(self.listener.check_delete_complete(None))
        self.assertFalse(self.listener._delete_called)
        self.octavia_client.listener_delete.assert_called_with('1234')
        self.assertFalse(self.listener.check_delete_complete(None))
        self.assertTrue(self.listener._delete_called)
        self.octavia_client.listener_delete.assert_called_with('1234')
        self.assertTrue(self.listener.check_delete_complete(None))

    def test_delete_not_found(self):
        self._create_stack()
        self.listener.resource_id_set('1234')
        self.octavia_client.listener_show.side_effect = [
            {'provisioning_status': 'PENDING_DELETE'},
        ]
        self.octavia_client.listener_delete.side_effect = [
            exceptions.Conflict(409),
            exceptions.NotFound(404)]

        self.listener.handle_delete()

        self.assertFalse(self.listener.check_delete_complete(None))
        self.assertFalse(self.listener._delete_called)
        self.octavia_client.listener_delete.assert_called_with('1234')
        self.assertTrue(self.listener.check_delete_complete(None))
        self.octavia_client.listener_delete.assert_called_with('1234')

    def test_delete_failed(self):
        self._create_stack()
        self.listener.resource_id_set('1234')
        self.octavia_client.listener_delete.side_effect = (
            exceptions.Unauthorized(401))

        self.listener.handle_delete()
        self.assertRaises(exceptions.Unauthorized,
                          self.listener.check_delete_complete, None)



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common import exception
from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support


class KeystoneRoleAssignmentMixin(object):
    """Implements role assignments between user/groups and project/domain.

    For example::

        heat_template_version: 2013-05-23

        parameters:
          ... Group or User parameters
          group_role:
            type: string
            description: role
          group_role_domain:
            type: string
            description: group role domain
          group_role_project:
            type: string
            description: group role project

        resources:
          admin_group:
            type: OS::Keystone::Group OR OS::Keystone::User
            properties:
              ... Group or User properties
              roles:
                - role: {get_param: group_role}
                  domain: {get_param: group_role_domain}
                - role: {get_param: group_role}
                  project: {get_param: group_role_project}
    """

    PROPERTIES = (
        ROLES
    ) = (
        'roles'
    )

    _ROLES_MAPPING_PROPERTIES = (
        ROLE, DOMAIN, PROJECT
    ) = (
        'role', 'domain', 'project'
    )

    mixin_properties_schema = {
        ROLES: properties.Schema(
            properties.Schema.LIST,
            _('List of role assignments.'),
            schema=properties.Schema(
                properties.Schema.MAP,
                _('Map between role with either project or domain.'),
                schema={
                    ROLE: properties.Schema(
                        properties.Schema.STRING,
                        _('Keystone role.'),
                        required=True,
                        constraints=([constraints.
                                     CustomConstraint('keystone.role')])
                    ),
                    PROJECT: properties.Schema(
                        properties.Schema.STRING,
                        _('Keystone project.'),
                        constraints=([constraints.
                                     CustomConstraint('keystone.project')])
                    ),
                    DOMAIN: properties.Schema(
                        properties.Schema.STRING,
                        _('Keystone domain.'),
                        constraints=([constraints.
                                     CustomConstraint('keystone.domain')])
                    ),
                }
            ),
            update_allowed=True
        )
    }

    def _add_role_assignments_to_group(self, group_id, role_assignments):
        for role_assignment in self._normalize_to_id(role_assignments):
            if role_assignment.get(self.PROJECT) is not None:
                self.client().roles.grant(
                    role=role_assignment.get(self.ROLE),
                    project=role_assignment.get(self.PROJECT),
                    group=group_id
                )
            elif role_assignment.get(self.DOMAIN) is not None:
                self.client().roles.grant(
                    role=role_assignment.get(self.ROLE),
                    domain=role_assignment.get(self.DOMAIN),
                    group=group_id
                )

    def _add_role_assignments_to_user(self, user_id, role_assignments):
        for role_assignment in self._normalize_to_id(role_assignments):
            if role_assignment.get(self.PROJECT) is not None:
                self.client().roles.grant(
                    role=role_assignment.get(self.ROLE),
                    project=role_assignment.get(self.PROJECT),
                    user=user_id
                )
            elif role_assignment.get(self.DOMAIN) is not None:
                self.client().roles.grant(
                    role=role_assignment.get(self.ROLE),
                    domain=role_assignment.get(self.DOMAIN),
                    user=user_id
                )

    def _remove_role_assignments_from_group(self, group_id, role_assignments,
                                            current_assignments):
        for role_assignment in self._normalize_to_id(role_assignments):
            if role_assignment in current_assignments:
                if role_assignment.get(self.PROJECT) is not None:
                    self.client().roles.revoke(
                        role=role_assignment.get(self.ROLE),
                        project=role_assignment.get(self.PROJECT),
                        group=group_id
                    )
                elif role_assignment.get(self.DOMAIN) is not None:
                    self.client().roles.revoke(
                        role=role_assignment.get(self.ROLE),
                        domain=role_assignment.get(self.DOMAIN),
                        group=group_id
                    )

    def _remove_role_assignments_from_user(self, user_id, role_assignments,
                                           current_assignments):
        for role_assignment in self._normalize_to_id(role_assignments):
            if role_assignment in current_assignments:
                if role_assignment.get(self.PROJECT) is not None:
                    self.client().roles.revoke(
                        role=role_assignment.get(self.ROLE),
                        project=role_assignment.get(self.PROJECT),
                        user=user_id
                    )
                elif role_assignment.get(self.DOMAIN) is not None:
                    self.client().roles.revoke(
                        role=role_assignment.get(self.ROLE),
                        domain=role_assignment.get(self.DOMAIN),
                        user=user_id
                    )

    def _normalize_to_id(self, role_assignment_prps):
        role_assignments = []
        if role_assignment_prps is None:
            return role_assignments

        for role_assignment in role_assignment_prps:
            role = role_assignment.get(self.ROLE)
            project = role_assignment.get(self.PROJECT)
            domain = role_assignment.get(self.DOMAIN)

            role_assignments.append({
                self.ROLE: self.client_plugin().get_role_id(role),
                self.PROJECT: (self.client_plugin().
                               get_project_id(project)) if project else None,
                self.DOMAIN: (self.client_plugin().
                              get_domain_id(domain)) if domain else None
            })
        return role_assignments

    def _find_differences(self, updated_prps, stored_prps):
        updated_role_project_assignments = []
        updated_role_domain_assignments = []

        # Split the properties into two set of role assignments
        # (project, domain) from updated properties
        for role_assignment in updated_prps or []:
            if role_assignment.get(self.PROJECT) is not None:
                updated_role_project_assignments.append(
                    '%s:%s' % (
                        role_assignment[self.ROLE],
                        role_assignment[self.PROJECT]))
            elif (role_assignment.get(self.DOMAIN)
                  is not None):
                updated_role_domain_assignments.append(
                    '%s:%s' % (role_assignment[self.ROLE],
                               role_assignment[self.DOMAIN]))

        stored_role_project_assignments = []
        stored_role_domain_assignments = []

        # Split the properties into two set of role assignments
        # (project, domain) from updated properties
        for role_assignment in (stored_prps or []):
            if role_assignment.get(self.PROJECT) is not None:
                stored_role_project_assignments.append(
                    '%s:%s' % (
                        role_assignment[self.ROLE],
                        role_assignment[self.PROJECT]))
            elif (role_assignment.get(self.DOMAIN)
                  is not None):
                stored_role_domain_assignments.append(
                    '%s:%s' % (role_assignment[self.ROLE],
                               role_assignment[self.DOMAIN]))

        new_role_assignments = []
        removed_role_assignments = []
        # NOTE: finding the diff of list of strings is easier by using 'set'
        #       so properties are converted to string in above sections
        # New items
        for item in (set(updated_role_project_assignments) -
                     set(stored_role_project_assignments)):
            new_role_assignments.append(
                {self.ROLE: item[:item.find(':')],
                 self.PROJECT: item[item.find(':') + 1:]}
            )

        for item in (set(updated_role_domain_assignments) -
                     set(stored_role_domain_assignments)):
            new_role_assignments.append(
                {self.ROLE: item[:item.find(':')],
                 self.DOMAIN: item[item.find(':') + 1:]}
            )

        # Old items
        for item in (set(stored_role_project_assignments) -
                     set(updated_role_project_assignments)):
            removed_role_assignments.append(
                {self.ROLE: item[:item.find(':')],
                 self.PROJECT: item[item.find(':') + 1:]}
            )
        for item in (set(stored_role_domain_assignments) -
                     set(updated_role_domain_assignments)):
            removed_role_assignments.append(
                {self.ROLE: item[:item.find(':')],
                 self.DOMAIN: item[item.find(':') + 1:]}
            )

        return new_role_assignments, removed_role_assignments

    def create_assignment(self, user_id=None, group_id=None):
        if self.properties.get(self.ROLES) is not None:
            if user_id is not None:
                self._add_role_assignments_to_user(
                    user_id,
                    self.properties.get(self.ROLES))
            elif group_id is not None:
                self._add_role_assignments_to_group(
                    group_id,
                    self.properties.get(self.ROLES))

    def update_assignment(self, prop_diff, user_id=None, group_id=None):
        # if there is no change do not update
        if self.ROLES in prop_diff:
            (new_role_assignments,
             removed_role_assignments) = self._find_differences(
                prop_diff.get(self.ROLES),
                self.properties[self.ROLES])

            if len(new_role_assignments) > 0:
                if user_id is not None:
                    self._add_role_assignments_to_user(
                        user_id,
                        new_role_assignments)
                elif group_id is not None:
                    self._add_role_assignments_to_group(
                        group_id,
                        new_role_assignments)

            if len(removed_role_assignments) > 0:
                current_assignments = self.parse_list_assignments(
                    user_id=user_id, group_id=group_id)
                if user_id is not None:
                    self._remove_role_assignments_from_user(
                        user_id,
                        removed_role_assignments,
                        current_assignments)
                elif group_id is not None:
                    self._remove_role_assignments_from_group(
                        group_id,
                        removed_role_assignments,
                        current_assignments)

    def delete_assignment(self, user_id=None, group_id=None):
        if self.properties[self.ROLES] is not None:
            current_assignments = self.parse_list_assignments(
                user_id=user_id, group_id=group_id)
            if user_id is not None:
                self._remove_role_assignments_from_user(
                    user_id,
                    (self.properties[self.ROLES]),
                    current_assignments)
            elif group_id is not None:
                self._remove_role_assignments_from_group(
                    group_id,
                    (self.properties[self.ROLES]),
                    current_assignments)

    def validate_assignment_properties(self):
        if self.properties.get(self.ROLES) is not None:
            for role_assignment in self.properties.get(self.ROLES):
                project = role_assignment.get(self.PROJECT)
                domain = role_assignment.get(self.DOMAIN)

                if project is not None and domain is not None:
                    raise exception.ResourcePropertyConflict(self.PROJECT,
                                                             self.DOMAIN)

                if project is None and domain is None:
                    msg = _('Either project or domain must be specified for'
                            ' role %s') % role_assignment.get(self.ROLE)
                    raise exception.StackValidationFailed(message=msg)

    def parse_list_assignments(self, user_id=None, group_id=None):
        """Method used for get_live_state implementation in other resources."""
        assignments = []
        roles = []
        if user_id is not None:
            assignments = self.client().role_assignments.list(user=user_id)
        elif group_id is not None:
            assignments = self.client().role_assignments.list(group=group_id)
        for assignment in assignments:
            values = assignment.to_dict()
            if not values.get('role') or not values.get('role').get('id'):
                continue
            role = {
                self.ROLE: values['role']['id'],
                self.DOMAIN: (values.get('scope') and
                              values['scope'].get('domain') and
                              values['scope'].get('domain').get('id')),
                self.PROJECT: (values.get('scope') and
                               values['scope'].get('project') and
                               values['scope'].get('project').get('id')),
            }
            roles.append(role)
        return roles


class KeystoneUserRoleAssignment(resource.Resource,
                                 KeystoneRoleAssignmentMixin):
    """Resource for granting roles to a user.

    Resource for specifying users and their's roles.
    """

    support_status = support.SupportStatus(
        version='5.0.0',
        message=_('Supported versions: keystone v3'))

    default_client_name = 'keystone'

    PROPERTIES = (
        USER,
    ) = (
        'user',
    )

    properties_schema = {
        USER: properties.Schema(
            properties.Schema.STRING,
            _('Name or id of keystone user.'),
            required=True,
            update_allowed=True,
            constraints=[constraints.CustomConstraint('keystone.user')]
        )
    }

    properties_schema.update(
        KeystoneRoleAssignmentMixin.mixin_properties_schema)

    def client(self):
        return super(KeystoneUserRoleAssignment, self).client().client

    @property
    def user_id(self):
        try:
            return self.client_plugin().get_user_id(
                self.properties.get(self.USER))
        except Exception as ex:
            self.client_plugin().ignore_not_found(ex)
            return None

    def handle_create(self):
        self.create_assignment(user_id=self.user_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self.update_assignment(user_id=self.user_id, prop_diff=prop_diff)

    def handle_delete(self):
        with self.client_plugin().ignore_not_found:
            self.delete_assignment(user_id=self.user_id)

    def validate(self):
        super(KeystoneUserRoleAssignment, self).validate()
        self.validate_assignment_properties()


class KeystoneGroupRoleAssignment(resource.Resource,
                                  KeystoneRoleAssignmentMixin):
    """Resource for granting roles to a group.

    Resource for specifying groups and their's roles.
    """

    support_status = support.SupportStatus(
        version='5.0.0',
        message=_('Supported versions: keystone v3'))

    default_client_name = 'keystone'

    PROPERTIES = (
        GROUP,
    ) = (
        'group',
    )

    properties_schema = {
        GROUP: properties.Schema(
            properties.Schema.STRING,
            _('Name or id of keystone group.'),
            required=True,
            update_allowed=True,
            constraints=[constraints.CustomConstraint('keystone.group')]
        )
    }

    properties_schema.update(
        KeystoneRoleAssignmentMixin.mixin_properties_schema)

    def client(self):
        return super(KeystoneGroupRoleAssignment, self).client().client

    @property
    def group_id(self):
        try:
            return self.client_plugin().get_group_id(
                self.properties.get(self.GROUP))
        except Exception as ex:
            self.client_plugin().ignore_not_found(ex)
            return None

    def handle_create(self):
        self.create_assignment(group_id=self.group_id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        self.update_assignment(group_id=self.group_id, prop_diff=prop_diff)

    def handle_delete(self):
        with self.client_plugin().ignore_not_found:
            self.delete_assignment(group_id=self.group_id)

    def validate(self):
        super(KeystoneGroupRoleAssignment, self).validate()
        self.validate_assignment_properties()


def resource_mapping():
    return {
        'OS::Keystone::UserRoleAssignment': KeystoneUserRoleAssignment,
        'OS::Keystone::GroupRoleAssignment': KeystoneGroupRoleAssignment
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging

from heat.common import exception
from heat.common.i18n import _
from heat.engine import attributes
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from heat.engine import translation

LOG = logging.getLogger(__name__)


class TroveCluster(resource.Resource):
    """A resource for managing Trove clusters.

    A Cluster is an opaque cluster used to store Database clusters.
    """

    support_status = support.SupportStatus(version='2015.1')

    TROVE_STATUS = (
        ERROR, FAILED, ACTIVE,
    ) = (
        'ERROR', 'FAILED', 'ACTIVE',
    )

    DELETE_STATUSES = (
        DELETING, NONE
    ) = (
        'DELETING', 'NONE'
    )

    TROVE_STATUS_REASON = {
        FAILED: _('The database instance was created, but heat failed to set '
                  'up the datastore. If a database instance is in the FAILED '
                  'state, it should be deleted and a new one should be '
                  'created.'),
        ERROR: _('The last operation for the database instance failed due to '
                 'an error.'),
    }

    BAD_STATUSES = (ERROR, FAILED)

    PROPERTIES = (
        NAME, DATASTORE_TYPE, DATASTORE_VERSION, INSTANCES,
    ) = (
        'name', 'datastore_type', 'datastore_version', 'instances',
    )

    _INSTANCE_KEYS = (
        FLAVOR, VOLUME_SIZE, NETWORKS, AVAILABILITY_ZONE,
    ) = (
        'flavor', 'volume_size', 'networks', 'availability_zone',
    )

    _NICS_KEYS = (
        NET, PORT, V4_FIXED_IP
    ) = (
        'network', 'port', 'fixed_ip'
    )

    ATTRIBUTES = (
        INSTANCES_ATTR, IP
    ) = (
        'instances', 'ip'
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of the cluster to create.'),
            constraints=[
                constraints.Length(max=255),
            ]
        ),
        DATASTORE_TYPE: properties.Schema(
            properties.Schema.STRING,
            _("Name of registered datastore type."),
            required=True,
            constraints=[
                constraints.Length(max=255)
            ]
        ),
        DATASTORE_VERSION: properties.Schema(
            properties.Schema.STRING,
            _("Name of the registered datastore version. "
              "It must exist for provided datastore type. "
              "Defaults to using single active version. "
              "If several active versions exist for provided datastore type, "
              "explicit value for this parameter must be specified."),
            required=True,
            constraints=[constraints.Length(max=255)]
        ),
        INSTANCES: properties.Schema(
            properties.Schema.LIST,
            _("List of database instances."),
            required=True,
            schema=properties.Schema(
                properties.Schema.MAP,
                schema={
                    FLAVOR: properties.Schema(
                        properties.Schema.STRING,
                        _('Flavor of the instance.'),
                        required=True,
                        constraints=[
                            constraints.CustomConstraint('trove.flavor')
                        ]
                    ),
                    VOLUME_SIZE: properties.Schema(
                        properties.Schema.INTEGER,
                        _('Size of the instance disk volume in GB.'),
                        required=True,
                        constraints=[
                            constraints.Range(1, 150),
                        ]
                    ),
                    NETWORKS: properties.Schema(
                        properties.Schema.LIST,
                        _("List of network interfaces to create on instance."),
                        support_status=support.SupportStatus(version='10.0.0'),
                        default=[],
                        schema=properties.Schema(
                            properties.Schema.MAP,
                            schema={
                                NET: properties.Schema(
                                    properties.Schema.STRING,
                                    _('Name or UUID of the network to attach '
                                      'this NIC to. Either %(port)s or '
                                      '%(net)s must be specified.') % {
                                        'port': PORT, 'net': NET},
                                    constraints=[
                                        constraints.CustomConstraint(
                                            'neutron.network')
                                    ]
                                ),
                                PORT: properties.Schema(
                                    properties.Schema.STRING,
                                    _('Name or UUID of Neutron port to '
                                      'attach this NIC to. Either %(port)s '
                                      'or %(net)s must be specified.')
                                    % {'port': PORT, 'net': NET},
                                    constraints=[
                                        constraints.CustomConstraint(
                                            'neutron.port')
                                    ],
                                ),
                                V4_FIXED_IP: properties.Schema(
                                    properties.Schema.STRING,
                                    _('Fixed IPv4 address for this NIC.'),
                                    constraints=[
                                        constraints.CustomConstraint('ip_addr')
                                    ]
                                ),
                            },
                        ),
                    ),
                    AVAILABILITY_ZONE: properties.Schema(
                        properties.Schema.STRING,
                        _('Name of the availability zone for DB instance.'),
                        support_status=support.SupportStatus(version='14.0.0'),
                    ),
                }
            )
        ),
    }

    attributes_schema = {
        INSTANCES: attributes.Schema(
            _("A list of instances ids."),
            type=attributes.Schema.LIST
        ),
        IP: attributes.Schema(
            _("A list of cluster instance IPs."),
            type=attributes.Schema.LIST
        )
    }

    default_client_name = 'trove'

    entity = 'clusters'

    def translation_rules(self, properties):
        return [
            translation.TranslationRule(
                properties,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.INSTANCES, self.NETWORKS, self.NET],
                client_plugin=self.client_plugin('neutron'),
                finder='find_resourceid_by_name_or_id',
                entity='network'),
            translation.TranslationRule(
                properties,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.INSTANCES, self.NETWORKS, self.PORT],
                client_plugin=self.client_plugin('neutron'),
                finder='find_resourceid_by_name_or_id',
                entity='port'),
            translation.TranslationRule(
                properties,
                translation.TranslationRule.RESOLVE,
                translation_path=[self.INSTANCES, self.FLAVOR],
                client_plugin=self.client_plugin(),
                finder='find_flavor_by_name_or_id'),
        ]

    def _cluster_name(self):
        return self.properties[self.NAME] or self.physical_resource_name()

    def handle_create(self):
        datastore_type = self.properties[self.DATASTORE_TYPE]
        datastore_version = self.properties[self.DATASTORE_VERSION]

        # convert instances to format required by troveclient
        instances = []
        for instance in self.properties[self.INSTANCES]:
            instance_dict = {
                'flavorRef': instance[self.FLAVOR],
                'volume': {'size': instance[self.VOLUME_SIZE]},
            }
            instance_nics = self.get_instance_nics(instance)
            if instance_nics:
                instance_dict["nics"] = instance_nics
            instance_availability_zone = instance[self.AVAILABILITY_ZONE]
            if instance_availability_zone:
                instance_dict["availability_zone"] = instance_availability_zone
            instances.append(instance_dict)

        args = {
            'name': self._cluster_name(),
            'datastore': datastore_type,
            'datastore_version': datastore_version,
            'instances': instances
        }
        cluster = self.client().clusters.create(**args)
        self.resource_id_set(cluster.id)
        return cluster.id

    def get_instance_nics(self, instance):
        nics = []
        for nic in instance[self.NETWORKS]:
            nic_dict = {}
            if nic.get(self.NET):
                nic_dict['net-id'] = nic.get(self.NET)
            if nic.get(self.PORT):
                nic_dict['port-id'] = nic.get(self.PORT)
            ip = nic.get(self.V4_FIXED_IP)
            if ip:
                nic_dict['v4-fixed-ip'] = ip
            nics.append(nic_dict)

        return nics

    def _refresh_cluster(self, cluster_id):
        try:
            cluster = self.client().clusters.get(cluster_id)
            return cluster
        except Exception as exc:
            if self.client_plugin().is_over_limit(exc):
                LOG.warning("Stack %(name)s (%(id)s) received an "
                            "OverLimit response during clusters.get():"
                            " %(exception)s",
                            {'name': self.stack.name,
                             'id': self.stack.id,
                             'exception': exc})
                return None
            else:
                raise

    def check_create_complete(self, cluster_id):
        cluster = self._refresh_cluster(cluster_id)

        if cluster is None:
            return False

        for instance in cluster.instances:
            if instance['status'] in self.BAD_STATUSES:
                raise exception.ResourceInError(
                    resource_status=instance['status'],
                    status_reason=self.TROVE_STATUS_REASON.get(
                        instance['status'], _("Unknown")))

            if instance['status'] != self.ACTIVE:
                return False

        LOG.info("Cluster '%s' has been created", cluster.name)
        return True

    def cluster_delete(self, cluster_id):
        try:
            cluster = self.client().clusters.get(cluster_id)
            cluster_status = cluster.task['name']
            if cluster_status not in self.DELETE_STATUSES:
                return False
            if cluster_status != self.DELETING:
                # If cluster already started to delete, don't send another one
                # request for deleting.
                cluster.delete()
        except Exception as ex:
            self.client_plugin().ignore_not_found(ex)
        return True

    def handle_delete(self):
        if not self.resource_id:
            return

        try:
            cluster = self.client().clusters.get(self.resource_id)
        except Exception as ex:
            self.client_plugin().ignore_not_found(ex)
        else:
            return cluster.id

    def check_delete_complete(self, cluster_id):
        if not cluster_id:
            return True

        if not self.cluster_delete(cluster_id):
            return False

        try:
            # For some time trove cluster may continue to live
            self._refresh_cluster(cluster_id)
        except Exception as ex:
            self.client_plugin().ignore_not_found(ex)
            return True

        return False

    def validate(self):
        res = super(TroveCluster, self).validate()
        if res:
            return res

        datastore_type = self.properties[self.DATASTORE_TYPE]
        datastore_version = self.properties[self.DATASTORE_VERSION]

        self.client_plugin().validate_datastore(
            datastore_type, datastore_version,
            self.DATASTORE_TYPE, self.DATASTORE_VERSION)

        for instance in self.properties[self.INSTANCES]:
            for nic in instance[self.NETWORKS]:
                if (nic.get(self.NET) is None and
                        nic.get(self.PORT) is None):
                    msg = (_("Either %(net)s or %(port)s must be provided.")
                           % {'net': self.NET, 'port': self.PORT})
                    raise exception.StackValidationFailed(message=msg)

                if (nic.get(self.NET) is not None and
                        nic.get(self.PORT) is not None):
                    raise exception.ResourcePropertyConflict(
                        self.NET, self.PORT)

    def _resolve_attribute(self, name):
        if self.resource_id is None:
            return
        if name == self.INSTANCES_ATTR:
            instances = []
            cluster = self.client().clusters.get(self.resource_id)
            for instance in cluster.instances:
                instances.append(instance['id'])
            return instances
        elif name == self.IP:
            cluster = self.client().clusters.get(self.resource_id)
            return cluster.ip


def resource_mapping():
    return {
        'OS::Trove::Cluster': TroveCluster
    }



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

from heat.common import template_format
from heat.engine.resources.openstack.octavia import availability_zone_profile
from heat.tests import common
from heat.tests.openstack.octavia import inline_templates
from heat.tests import utils


class AvailabilityZoneProfileTest(common.HeatTestCase):

    def test_resource_mapping(self):
        mapping = availability_zone_profile.resource_mapping()
        self.assertEqual(
            availability_zone_profile.AvailabilityZoneProfile,
            mapping['OS::Octavia::AvailabilityZoneProfile'],
        )

    def _create_stack(
        self, tmpl=inline_templates.AVAILABILITY_ZONE_PROFILE_TEMPLATE
    ):
        self.t = template_format.parse(tmpl)
        self.stack = utils.parse_stack(self.t)
        self.az = self.stack['availability_zone_profile']

        self.octavia_client = mock.MagicMock()
        self.az.client = mock.MagicMock()
        self.az.client.return_value = self.octavia_client

        self.az.resource_id_set('1234')
        self.patchobject(
            self.az, 'physical_resource_name', return_value='resource_name'
        )

    def test_create(self):
        self._create_stack()
        expected = {
            'availability_zone_profile': {
                'name': 'test_availability_zone_profile',
                'availability_zone_data': '{"compute_zone": "az-central"}',
                'provider_name': 'amphora',
            }
        }

        self.az.handle_create()

        self.octavia_client.availabilityzoneprofile_create.assert_called_with(
            json=expected
        )

    def test_show_resource(self):
        self._create_stack()
        self.octavia_client.availabilityzoneprofile_show.return_value = {
            'id': 'azp_id_1234'
        }
        self.assertEqual({'id': 'azp_id_1234'}, self.az._show_resource())

        self.octavia_client.availabilityzoneprofile_show.assert_called_with(
            '1234'
        )

    def test_update(self):
        self._create_stack()
        prop_diff = {
            'name': 'test_availability_zone_profile2',
            'availability_zone_data': '{"compute_zone": "az-edge2"}',
            'provider_name': 'amphora2',
        }

        self.az.handle_update(None, None, prop_diff)

        octavia_client = self.octavia_client
        octavia_client.availabilityzoneprofile_set.assert_called_once_with(
            '1234', json={'availability_zone_profile': prop_diff}
        )

        octavia_client.availabilityzoneprofile_set.reset_mock()

        # Updating an availability zone with None as name should use
        # physical_resource_name() as new name
        prop_diff = {
            'name': None,
            'availability_zone_data': '{"compute_zone": "az-edge3"}',
            'provider_name': 'amphora3',
        }

        self.az.handle_update(None, None, prop_diff)

        self.assertEqual(prop_diff['name'], 'resource_name')
        octavia_client.availabilityzoneprofile_set.assert_called_once_with(
            '1234', json={'availability_zone_profile': prop_diff}
        )

    def test_delete(self):
        self._create_stack()

        self.az.handle_delete()

        self.octavia_client.availabilityzoneprofile_delete.assert_called_with(
            '1234'
        )



#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heat.common.i18n import _
from heat.engine import constraints
from heat.engine import properties
from heat.engine import resource
from heat.engine import support
from heat.engine import translation


class KeystoneEndpoint(resource.Resource):
    """Heat Template Resource for Keystone Service Endpoint.

    Keystone endpoint is just the URL that can be used for accessing a service
    within OpenStack. Endpoint can be accessed by admin, by services or public,
    i.e. everyone can use this endpoint.
    """

    support_status = support.SupportStatus(
        version='5.0.0',
        message=_('Supported versions: keystone v3'))

    default_client_name = 'keystone'

    entity = 'endpoints'

    PROPERTIES = (
        NAME, REGION, SERVICE, INTERFACE, SERVICE_URL, ENABLED,
    ) = (
        'name', 'region', 'service', 'interface', 'url', 'enabled',
    )

    properties_schema = {
        NAME: properties.Schema(
            properties.Schema.STRING,
            _('Name of keystone endpoint.'),
            update_allowed=True
        ),
        REGION: properties.Schema(
            properties.Schema.STRING,
            _('Name or Id of keystone region.'),
            update_allowed=True,
            constraints=[constraints.CustomConstraint('keystone.region')]
        ),
        SERVICE: properties.Schema(
            properties.Schema.STRING,
            _('Name or Id of keystone service.'),
            update_allowed=True,
            required=True,
            constraints=[constraints.CustomConstraint('keystone.service')]
        ),
        INTERFACE: properties.Schema(
            properties.Schema.STRING,
            _('Interface type of keystone service endpoint.'),
            update_allowed=True,
            required=True,
            constraints=[constraints.AllowedValues(
                ['public', 'internal', 'admin']
            )]
        ),
        SERVICE_URL: properties.Schema(
            properties.Schema.STRING,
            _('URL of keystone service endpoint.'),
            update_allowed=True,
            required=True
        ),
        ENABLED: properties.Schema(
            properties.Schema.BOOLEAN,
            _('This endpoint is enabled or disabled.'),
            default=True,
            update_allowed=True,
            support_status=support.SupportStatus(version='6.0.0')
        )
    }

    def translation_rules(self, props):
        return [
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.SERVICE],
                client_plugin=self.client_plugin(),
                finder='get_service_id'
            ),
            translation.TranslationRule(
                props,
                translation.TranslationRule.RESOLVE,
                [self.REGION],
                client_plugin=self.client_plugin(),
                finder='get_region_id'
            ),
        ]

    def client(self):
        return super(KeystoneEndpoint, self).client().client

    def handle_create(self):
        region = self.properties[self.REGION]
        service = self.properties[self.SERVICE]
        interface = self.properties[self.INTERFACE]
        url = self.properties[self.SERVICE_URL]
        name = (self.properties[self.NAME] or
                self.physical_resource_name())
        enabled = self.properties[self.ENABLED]

        endpoint = self.client().endpoints.create(
            region=region,
            service=service,
            interface=interface,
            url=url,
            name=name,
            enabled=enabled)

        self.resource_id_set(endpoint.id)

    def handle_update(self, json_snippet, tmpl_diff, prop_diff):
        if prop_diff:
            region = prop_diff.get(self.REGION)
            service = prop_diff.get(self.SERVICE)
            interface = prop_diff.get(self.INTERFACE)
            url = prop_diff.get(self.SERVICE_URL)
            name = None
            # Don't update the name if no change
            if self.NAME in prop_diff:
                name = prop_diff[self.NAME] or self.physical_resource_name()
            enabled = prop_diff.get(self.ENABLED)

            self.client().endpoints.update(
                endpoint=self.resource_id,
                region=region,
                service=service,
                interface=interface,
                url=url,
                name=name,
                enabled=enabled)

    def parse_live_resource_data(self, resource_properties, resource_data):
        endpoint_reality = {}

        endpoint_reality.update(
            {self.SERVICE: resource_data.get('service_id'),
             self.REGION: resource_data.get('region_id')})
        for key in (set(self.PROPERTIES) - {self.SERVICE, self.REGION}):
            endpoint_reality.update({key: resource_data.get(key)})
        return endpoint_reality


def resource_mapping():
    return {
        'OS::Keystone::Endpoint': KeystoneEndpoint
    }
